import unittest
from pyERGM.utils import *
from pyERGM.metrics import *
from pyERGM.ergm import ERGM, BruteForceERGM
from pyERGM.datasets import sampson_matrix
import sys
from scipy.linalg import eigh


class TestERGM(unittest.TestCase):
    def setUp(self):
        self.metrics = [NumberOfEdgesUndirected(), NumberOfTriangles()]
        self.n_nodes = 3

        self.K = 100
        self.thetas = {str(m): 1 for m in self.metrics}

    def test_calculate_weight(self):
        ergm = ERGM(self.n_nodes, self.metrics, is_directed=False, initial_thetas=self.thetas,
                    initial_normalization_factor=self.K)

        W = np.array([[0, 1, 1],
                      [1, 0, 1],
                      [1, 1, 0]])
        weight = ergm.calculate_weight(W)

        expected_num_edges = 3
        expected_num_triangles = 1
        expected_weight = np.exp(expected_num_edges * 1 + expected_num_triangles * 1)

        self.assertEqual(weight, expected_weight)

        W = np.array([[0, 0, 0],
                      [0, 0, 0],
                      [0, 0, 0]])
        weight = ergm.calculate_weight(W)

        expected_num_edges = 0
        expected_num_triangles = 0
        expected_weight = np.exp(expected_num_edges * 1 + expected_num_triangles * 1)

        self.assertEqual(weight, expected_weight)

    def test_calculate_probability(self):
        ergm = ERGM(self.n_nodes, self.metrics, is_directed=False, initial_thetas=self.thetas,
                    initial_normalization_factor=self.K)

        W = np.array([[0, 1, 1],
                      [1, 0, 1],
                      [1, 1, 0]])
        probability = ergm.calculate_probability(W)

        expected_num_edges = 3
        expected_num_triangles = 1
        expected_weight = np.exp(expected_num_edges * 1 + expected_num_triangles * 1)
        expected_probability = expected_weight / self.K

        self.assertEqual(probability, expected_probability)

    def test_calculate_probability_wiki_example(self):
        thetas = [-np.log(2), np.log(3)]
        K = 29 / 8

        ergm = ERGM(self.n_nodes, self.metrics, is_directed=False,
                    initial_thetas={str(m): thetas[i] for i, m in enumerate(self.metrics)},
                    initial_normalization_factor=K)

        W_0_edges = np.array([[0, 0, 0],
                              [0, 0, 0],
                              [0, 0, 0]])

        probability = ergm.calculate_probability(W_0_edges)
        expected_probability = 1 / K
        self.assertEqual(probability, expected_probability)

        W_1_edges = np.array([[0, 0, 1],
                              [0, 0, 0],
                              [1, 0, 0]])

        probability = ergm.calculate_probability(W_1_edges)
        expected_probability = 0.5 / K
        self.assertEqual(probability, expected_probability)

        W_2_edges = np.array([[0, 1, 1],
                              [1, 0, 0],
                              [1, 0, 0]])

        probability = ergm.calculate_probability(W_2_edges)
        expected_probability = 0.25 / K
        self.assertEqual(probability, expected_probability)

        W_3_edges = np.array([[0, 1, 1],
                              [1, 0, 1],
                              [1, 1, 0]])

        probability = round(ergm.calculate_probability(W_3_edges), 6)
        expected_probability = round(0.375 / K, 6)
        self.assertEqual(probability, expected_probability)

    def test_benchmark_er_convergence(self, n=5, p=0.1, is_directed=False):
        np.random.seed(9873645)
        print(f"Running an ERGM bruteforce fit with {n} nodes, p={p}, directed={is_directed}")
        num_pos_connect = n * (n - 1)

        if not is_directed:
            num_pos_connect //= 2

        ground_truth_num_edges = round(num_pos_connect * p)
        ground_truth_p = ground_truth_num_edges / num_pos_connect
        ground_truth_theta = np.array([np.log(ground_truth_p / (1 - ground_truth_p))])

        adj_mat_no_diag = np.zeros(num_pos_connect)
        on_indices = np.random.choice(num_pos_connect, size=ground_truth_num_edges, replace=False).astype(int)
        adj_mat_no_diag[on_indices] = 1
        adj_mat = np.zeros((n, n))

        if not is_directed:
            upper_triangle_indices = np.triu_indices(n, k=1)
            adj_mat[upper_triangle_indices] = adj_mat_no_diag
            lower_triangle_indices_aligned = (upper_triangle_indices[1], upper_triangle_indices[0])
            adj_mat[lower_triangle_indices_aligned] = adj_mat_no_diag
        else:
            adj_mat[~np.eye(n, dtype=bool)] = adj_mat_no_diag

        number_of_edges_metric = NumberOfEdgesDirected() if is_directed else NumberOfEdgesUndirected()
        model = BruteForceERGM(n, [number_of_edges_metric], is_directed=is_directed)
        model.fit(adj_mat)

        print(f"ground truth theta: {ground_truth_theta}")
        print(f"fit theta: {model._thetas}")

        for t_model, t_ground_truth in zip(model._thetas, ground_truth_theta):
            self.assertAlmostEqual(t_model, t_ground_truth, places=5)

        non_synapses_indices = np.where(adj_mat_no_diag == 0)[0]
        prediction = ground_truth_p * np.ones(adj_mat_no_diag.size)
        prediction[non_synapses_indices] = 1 - ground_truth_p
        true_log_like = np.log(prediction).sum()
        print(f"true log likelihood: {true_log_like}")

        number_of_edges_metric = NumberOfEdgesDirected() if is_directed else NumberOfEdgesUndirected()
        model_with_true_theta = BruteForceERGM(n, [number_of_edges_metric],
                                               initial_thetas={str(number_of_edges_metric): ground_truth_theta[0]},
                                               is_directed=is_directed)

        ground_truth_model_log_like = np.log(model_with_true_theta.calculate_weight(adj_mat)) - np.log(
            model_with_true_theta._normalization_factor)

        print(f"model with true theta log like: {ground_truth_model_log_like}")
        print(f"normalization factor: {model_with_true_theta._normalization_factor}")

    def test_fit_small_ER_network(self):
        n = 4
        p = 0.25
        is_directed = False

        num_pos_connect = n * (n - 1)
        if not is_directed:
            num_pos_connect //= 2

        ground_truth_num_edges = round(num_pos_connect * p)
        ground_truth_p = ground_truth_num_edges / num_pos_connect
        ground_truth_theta = np.array([np.log(ground_truth_p / (1 - ground_truth_p))])

        adj_mat_no_diag = np.zeros(num_pos_connect)
        on_indices = np.random.choice(num_pos_connect, size=ground_truth_num_edges, replace=False).astype(int)
        adj_mat_no_diag[on_indices] = 1
        adj_mat = np.zeros((n, n))

        if not is_directed:
            upper_triangle_indices = np.triu_indices(n, k=1)
            adj_mat[upper_triangle_indices] = adj_mat_no_diag
            lower_triangle_indices_aligned = (upper_triangle_indices[1], upper_triangle_indices[0])
            adj_mat[lower_triangle_indices_aligned] = adj_mat_no_diag
        else:
            adj_mat[~np.eye(n, dtype=bool)] = adj_mat_no_diag

        W = adj_mat
        print("Fitting matrix - ")
        print(W)

        ergm = ERGM(n,
                    [NumberOfEdgesUndirected()],
                    is_directed=is_directed,
                    seed_MCMC_proba=0.25
                    )

        ergm.fit(W, lr=0.01, opt_steps=300, sliding_grad_window_k=10, sample_pct_growth=0.05, steps_for_decay=20,
                 lr_decay_pct=0.05, mcmc_sample_size=200, mcmc_steps_per_sample=10)

        fit_theta = ergm._thetas[0]
        print(f"ground truth theta: {ground_truth_theta}")
        print(f"fit theta: {fit_theta}")

        # TODO - what criteria to use for testing convergence? From manual tests, it doesn't seem to perfectly converge on the true thetas...
        # Nevertheless, even without an assert, this will catch errors - fit won't work if something breaks.

    def test_fit_small_ER_network_with_e_i_node_feature(self):
        n = 4
        p = 0.25
        is_directed = False

        num_pos_connect = n * (n - 1)
        if not is_directed:
            num_pos_connect //= 2

        ground_truth_num_edges = round(num_pos_connect * p)
        ground_truth_p = ground_truth_num_edges / num_pos_connect

        ground_truth_theta = np.log(ground_truth_p / (1 - ground_truth_p))

        adj_mat_no_diag = np.zeros(num_pos_connect)
        on_indices = np.random.choice(num_pos_connect, size=ground_truth_num_edges, replace=False).astype(int)
        adj_mat_no_diag[on_indices] = 1
        adj_mat = np.zeros((n, n))

        if not is_directed:
            upper_triangle_indices = np.triu_indices(n, k=1)
            adj_mat[upper_triangle_indices] = adj_mat_no_diag
            lower_triangle_indices_aligned = (upper_triangle_indices[1], upper_triangle_indices[0])
            adj_mat[lower_triangle_indices_aligned] = adj_mat_no_diag
        else:
            adj_mat[~np.eye(n, dtype=bool)] = adj_mat_no_diag

        node_features = {"E_I": [[0, 1, 1, 1]]}
        ground_truth_e_nodes = 0.25
        ground_truth_theta_e = np.log(ground_truth_e_nodes / (1 - ground_truth_e_nodes))

        W = adj_mat # np.concatenate([adj_mat, node_types], axis=1)
        print(f"Fitting matrix ({W.shape}) - ")
        print(W)

        ergm = ERGM(n,
                    [NumberOfEdgesUndirected(), NumberOfNodesPerType(metric_node_feature='E_I', n_node_categories=2)],
                    is_directed=is_directed,
                    seed_MCMC_proba=0.25
                    )

        ergm.fit(W, observed_node_features=node_features, lr=0.1, opt_steps=500, sliding_grad_window_k=10, sample_pct_growth=0.05,
                 steps_for_decay=20, lr_decay_pct=0.05, mcmc_sample_size=300, mcmc_steps_per_sample=10, no_mple=True,
                 theta_init_method='uniform')

        fit_theta = ergm._thetas[0]
        fit_theta_e = ergm._thetas[1]
        print(f"ground truth theta: {ground_truth_theta, ground_truth_theta_e}")
        print(f"fit theta: {fit_theta, fit_theta_e}")


        # ergm._thetas = np.array([ground_truth_theta, ground_truth_theta_e])
        sampled_networks = ergm.generate_networks_for_sample(sample_size=100)
        fit_p = sampled_networks[:, :-1, :].mean()
        fit_e_i_fraction = sampled_networks[:, -1, :].mean()
        print(f"ground truth p: {p}")
        print(f"fit p: {fit_p}")
        print(f"ground truth excitatory fraction: {np.array(node_features['E_I']).mean()}")
        print(f"fit excitatory fraction: {fit_e_i_fraction}")

    def test_MPLE(self):
        n = 4

        M1 = np.array([
            [0, 0, 0, 1],
            [1, 0, 0, 0],
            [0, 0, 0, 0],
            [1, 0, 1, 0]
        ])

        types = ["A", "A", "B", "B"]
        metrics = [NumberOfEdgesTypesDirected(types)]
        model = ERGM(n, metrics, is_directed=True)
        result = model.fit(M1, mple_lr=1)

        self.assertTrue(result["success"])

        inferred_probas_per_type_pairs = list(np.exp(model._thetas) / (1 + np.exp(model._thetas)))

        real_densities_per_type = get_edge_density_per_type_pairs(M1, types)
        real_densities_per_type = list(real_densities_per_type.values())

        for inferred_proba, real_density in zip(inferred_probas_per_type_pairs, real_densities_per_type):
            self.assertAlmostEqual(inferred_proba, real_density, places=4)

    def test_MPLE_convergence_with_auto_LR_decay(self):
        n = 4

        M1 = np.array([
            [0, 0, 0, 1],
            [1, 0, 0, 0],
            [0, 0, 0, 0],
            [1, 0, 1, 0]
        ])

        types = ["A", "A", "B", "B"]
        metrics = [NumberOfEdgesTypesDirected(types)]
        model = ERGM(n, metrics, is_directed=True)
        model.fit(M1, mple_max_iter=1000, mple_lr=1000, mple_stopping_thr=1e-10)

        inferred_probas_per_type_pairs = list(np.exp(model._thetas) / (1 + np.exp(model._thetas)))

        real_densities_per_type = get_edge_density_per_type_pairs(M1, types)
        real_densities_per_type = list(real_densities_per_type.values())

        for inferred_proba, real_density in zip(inferred_probas_per_type_pairs, real_densities_per_type):
            self.assertAlmostEqual(inferred_proba, real_density, places=4)
    
    def test_sampson_MCMLE(self):
        np.random.seed(1234)
        metrics = [NumberOfEdgesDirected(), OutDegree(), InDegree(), TotalReciprocity()]
        n_nodes = sampson_matrix.shape[0]

        mcmle_model = ERGM(n_nodes, metrics, is_directed=True)

        convergence_result = mcmle_model.fit(sampson_matrix, 
                opt_steps=10, 
                steps_for_decay=1, 
                lr=1, 
                mple_lr=0.5, 
                convergence_criterion="model_bootstrap",
                mcmc_burn_in=0,
                mcmc_steps_per_sample=n_nodes**2,
                mcmc_sample_size=1000,
                num_model_sub_samples=10,
                model_subsample_size=1000,
                bootstrap_convergence_confidence=0.95,
                bootstrap_convergence_num_stds_away_thr=1,
            )
    
        model_thethas = mcmle_model._thetas

        expected_values = {"edges": -1.1761, "sender2": -0.2945, "sender3": 1.4141, "sender4": 0.3662, "sender5": 0.1315,
               "sender6": 1.2148, "sender7": 0.6055,
               "sender8": 1.3609, "sender9": 0.6402, "sender10": 2.0639, "sender11": 1.4355, "sender12": -0.1681,
               "sender13": -0.2322, "sender14": 0.5841, "sender15": 1.8600,
               "sender16": 1.4317, "sender17": 1.2211, "sender18": 1.8724, "receiver2": -0.1522, "receiver3": -3.0453,
               "receiver4": -1.7596, "receiver5": -0.8198, "receiver6": -3.3922,
               "receiver7": -1.6074, "receiver8": -2.2656, "receiver9": -2.2069, "receiver10": -3.9189,
               "receiver11": -3.0257, "receiver12": -0.9457, "receiver13": -1.4749, "receiver14": -1.5950,
               "receiver15": -3.3147, "receiver16": -3.0567, "receiver17": -3.4436, "receiver18": -3.3239,
               "mutual": 3.6918
               }
        expected_thetas = np.array(list(expected_values.values()))
        
        thetas_R_2 = 1 - np.sum((model_thethas - expected_thetas)**2) / np.sum((expected_thetas - np.mean(expected_thetas))**2)
        self.assertTrue(thetas_R_2 > 0.99)
        self.assertTrue(convergence_result["success"])

    def test_sampson_MCMLE_with_node_features(self):
        np.random.seed(1234)
        metrics = [NumberOfEdgesDirected(), OutDegree(), InDegree(), TotalReciprocity()]
        n_nodes = sampson_matrix.shape[0]

        mcmle_model = ERGM(n_nodes, metrics, is_directed=True)

        node_features = np.random.choice(3, size=(n_nodes, 2))
        sampson_matrix_with_node_features = np.concatenate([sampson_matrix, node_features], axis=1)
        convergence_result = mcmle_model.fit(sampson_matrix_with_node_features,
                                             opt_steps=10,
                                             steps_for_decay=1,
                                             lr=1,
                                             mple_lr=0.5,
                                             convergence_criterion="model_bootstrap",
                                             mcmc_burn_in=0,
                                             mcmc_steps_per_sample=n_nodes ** 2,
                                             mcmc_sample_size=1000,
                                             num_model_sub_samples=10,
                                             model_subsample_size=1000,
                                             bootstrap_convergence_confidence=0.95,
                                             bootstrap_convergence_num_stds_away_thr=1,
                                             )

        model_thethas = mcmle_model._thetas

        expected_values = {"edges": -1.1761, "sender2": -0.2945, "sender3": 1.4141, "sender4": 0.3662,
                           "sender5": 0.1315,
                           "sender6": 1.2148, "sender7": 0.6055,
                           "sender8": 1.3609, "sender9": 0.6402, "sender10": 2.0639, "sender11": 1.4355,
                           "sender12": -0.1681,
                           "sender13": -0.2322, "sender14": 0.5841, "sender15": 1.8600,
                           "sender16": 1.4317, "sender17": 1.2211, "sender18": 1.8724, "receiver2": -0.1522,
                           "receiver3": -3.0453,
                           "receiver4": -1.7596, "receiver5": -0.8198, "receiver6": -3.3922,
                           "receiver7": -1.6074, "receiver8": -2.2656, "receiver9": -2.2069, "receiver10": -3.9189,
                           "receiver11": -3.0257, "receiver12": -0.9457, "receiver13": -1.4749, "receiver14": -1.5950,
                           "receiver15": -3.3147, "receiver16": -3.0567, "receiver17": -3.4436, "receiver18": -3.3239,
                           "mutual": 3.6918
                           }
        expected_thetas = np.array(list(expected_values.values()))

        thetas_R_2 = 1 - np.sum((model_thethas - expected_thetas) ** 2) / np.sum(
            (expected_thetas - np.mean(expected_thetas)) ** 2)
        self.assertTrue(thetas_R_2 > 0.99)
        self.assertTrue(convergence_result["success"])

    def test_MPLE_regressors_of_different_scales(self):
        # TODO: currently this is a smoke test - we validate nothing: neither convergence nor the thetas/predictions.
        #  Somehow sklearn still finds a slightly better solution than ours.
        np.random.seed(42)

        W = np.random.randint(0, 2, size=(10, 10))
        W[np.diag_indices(10)] = 0

        metrics = [NumberOfEdgesDirected(),
                   NodeAttrSum([np.random.randint(1, 5) ** 10 for x in range(10)], is_directed=True)]
        model = ERGM(n_nodes=10, metrics_collection=metrics, is_directed=True)

        model.fit(W)

        # sklearn_thetas = np.array([2.82974701e-01, -2.34383474e-07])

        # sklearn_probas = np.array([0.56682151,
        # 0.56682151,
        # 0.50584116,
        # 0.56682151,
        # 0.56682151,
        # 0.56682151,
        # 0.50584116,
        # 0.56682151,
        # 0.56347922,
        # 0.56682151,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.57015772,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.56682151,
        # 0.56682151,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.57015772,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.56682151,
        # 0.50584116,
        # 0.5092404 ,
        # 0.5092404 ,
        # 0.5092404 ,
        # 0.5092404 ,
        # 0.5092404 ,
        # 0.44804742,
        # 0.5092404 ,
        # 0.50584116,
        # 0.56682151,
        # 0.57015772,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.56682151,
        # 0.56682151,
        # 0.57015772,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.56682151,
        # 0.56682151,
        # 0.57015772,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.56682151,
        # 0.50584116,
        # 0.5092404 ,
        # 0.5092404 ,
        # 0.44804742,
        # 0.5092404 ,
        # 0.5092404 ,
        # 0.5092404 ,
        # 0.5092404 ,
        # 0.50584116,
        # 0.56682151,
        # 0.57015772,
        # 0.57015772,
        # 0.5092404 ,
        # 0.57015772,
        # 0.57015772,
        # 0.57015772,
        # 0.5092404 ,
        # 0.56682151,
        # 0.56347922,
        # 0.56682151,
        # 0.56682151,
        # 0.50584116,
        # 0.56682151,
        # 0.56682151,
        # 0.56682151,
        # 0.50584116,
        # 0.56682151])

        # self.assertAlmostEqual()

    def test_assigning_model_initial_thetas(self):
        # TODO: seems like convergence of the model in this test depends on the seed...
        np.random.seed(8765)
        n_nodes = 5
        W = np.array([[0, 0, 1, 1, 0],
                      [1, 0, 0, 0, 1],
                      [0, 0, 0, 0, 1],
                      [0, 1, 0, 0, 0],
                      [0, 1, 0, 1, 0]])
        metrics_1 = [NumberOfEdgesDirected(), NodeAttrSum(np.arange(1, n_nodes + 1), is_directed=True),
                     NumberOfEdgesTypesDirected(['A', 'B', 'A', 'A', 'B'])]
        model_1 = ERGM(n_nodes=n_nodes, metrics_collection=metrics_1, is_directed=True)

        model_1.fit(W)

        model_1_params = model_1.get_model_parameters()

        metrics_2 = [NumberOfEdgesDirected(), NodeAttrSum(np.arange(n_nodes + 1, 1, -1), is_directed=True),
                     NumberOfEdgesTypesDirected(['B', 'B', 'B', 'A', 'A'])]
        model_2 = ERGM(n_nodes=n_nodes, metrics_collection=metrics_2, is_directed=True, initial_thetas=model_1_params)

        self.assertTrue(model_2.get_model_parameters() == model_1_params)

    def test_calculate_prediction(self):
        n_nodes = 4
        W = np.array([[0, 1, 0, 1],
                      [0, 0, 1, 1],
                      [1, 0, 0, 0],
                      [0, 0, 1, 0]])

        metrics = [NumberOfEdgesDirected()]
        model = ERGM(n_nodes=n_nodes, metrics_collection=metrics, is_directed=True)
        model.fit(W)

        model_2 = ERGM(n_nodes=n_nodes, metrics_collection=metrics, is_directed=True,
                       initial_thetas=model.get_model_parameters())
        model_2_av_mat = model_2.get_mple_prediction(W)
        expected_model_2_av_mat = 0.5 * np.ones((n_nodes, n_nodes))
        expected_model_2_av_mat[np.diag_indices(n_nodes)] = 0
        self.assertTrue(np.abs(model_2_av_mat - expected_model_2_av_mat).max() < 1e-10)
    
    def test_model_initialization_from_existing_params_(self):
        np.random.seed(1234)
        metrics = [NumberOfEdgesDirected(), OutDegree(), InDegree(), TotalReciprocity()]
        n_nodes = sampson_matrix.shape[0]

        mcmle_model = ERGM(n_nodes, metrics, is_directed=True)

        model_params = mcmle_model.get_model_parameters()

        new_model = ERGM(n_nodes, metrics, is_directed=True, initial_thetas=model_params)
        
        # If there's a problem with copying the parameters, this will throw an error.
        new_model.generate_networks_for_sample(sample_size=10)







from copy import deepcopy
import numpy as np
from pyERGM.utils import *
from pyERGM.metrics import MetricsCollection
import time
from scipy.special import softmax


class Sampler():
    def __init__(self, thetas, metrics_collection: MetricsCollection):
        self.thetas = deepcopy(thetas)
        self.metrics_collection = deepcopy(metrics_collection)

    def sample(self, initial_state, n_iterations):
        pass


class NaiveMetropolisHastings(Sampler):
    def __init__(self, thetas, metrics_collection: MetricsCollection):
        """
        An implementation for the symmetric proposal Metropolis-Hastings algorithm for ERGMS, using the logit
        of the acceptance rate. See docs for more details.
        Throughout this implementation, networks are represented as adjacency matrices.

        Parameters
        ----------
        thetas : np.ndarray
            Coefficients of the ERGM
            # TODO: should be a field of the class or passed to sample?
        
        metrics_collection : MetricsCollection
            A MetricsCollection object that can calculate statistics of a network.
        """
        super().__init__(thetas, metrics_collection)

        self._edge_proposal_dists = {}

    def set_thetas(self, thetas):
        self.thetas = deepcopy(thetas)

    def _calculate_weighted_change_score(self, current_network, edge_flip_info: dict, node_flip_info={}):
        """
        Calculate g(proposed_network)-g(current_network) and then inner product with thetas.
        """
        change_score = self.metrics_collection.calc_change_scores(current_network, edge_flip_info, node_flip_info)
        return np.dot(self.thetas, change_score)

    def _flip_network_edge(self, current_network, i, j):
        """
        Flip the edge between nodes i, j. If it's an undirected network, we flip entries W_i,j and W_j,i.
        NOTE! This function changes the network that is passed by reference
        """
        current_network[i, j] = 1 - current_network[i, j]
        if not self.metrics_collection.is_directed:
            current_network[j, i] = 1 - current_network[j, i]

    def _update_node_features(self, current_network, i, f, c):
        """
        Update node i's feature f (feature index) to be of category c.
        NOTE! This function changes the network that is passed by reference
        """
        num_nodes = current_network.shape[0]
        current_network[i, num_nodes + f] = c

    def _calc_edge_influence_on_features(self, net_for_change_scores: np.ndarray):
        """
        Calculates the influence of each edge on each one of the features in the MetricsCollection.
        The influence of an edge on a feature is defined as the ratio between the change in the feature that flipping
        the edge would cause divided by the average change over all edges.
        If the average influence is 0 (the feature does not change by flipping any edge), the influence of all edges is
        0.
        TODO: maybe we need to handle these all-0 cases separately (e.g., on the empty network, no edge flip will change
         the reciprocity, but obviously all edges have influence on the reciprocity). Or maybe this is not a good
         definition for the influence of an edge.
        For example, each edge has an influence of 1 on NumberOfEdges (no matter what edge we flip, it changes by 1).
        Similarly, an edge (i,j) has an influence of n on OutDegree of node i, and 0 on OutDegree of other nodes (the
        out degree of node 1 would change by 1 as a result of flipping the edge, and there are exactly n-1 such edges,
        out of the n(n-1) edges in the matrix, so the average influence is 1/n).
        Parameters
        ----------
        net_for_change_scores : np.ndarray
            The network using which the change scores for all edges is calculated.
        Returns
        -------
        A n(n-1) vector with the sum of influences of each edge over all features (total influence of each edge).
        """
        change_scores, _ = self.metrics_collection.prepare_mple_data(net_for_change_scores)
        mean_importance_per_feature = change_scores.mean(axis=0)
        change_scores[:, mean_importance_per_feature != 0] /= mean_importance_per_feature[
            None, mean_importance_per_feature != 0]
        return change_scores.sum(axis=1)

    def _calc_proposal_dist_features_influence__sum(self, net_for_change_scores: np.ndarray):
        """
        Calculate a proposal distribution over edges such that edges with high influence will get higher probabilities.
        The transformation of total influences into a distribution is done by normalized using the sum of total
        influences.
        Parameters
        ----------
        net_for_change_scores : np.ndarray
            The network using which the change scores for all edges is calculated.
        """
        edge_influence = self._calc_edge_influence_on_features(net_for_change_scores)
        self._edge_proposal_dists['features_influence__sum'] = edge_influence / edge_influence.sum()

    def _calc_proposal_dist_features_influence__softmax(self, net_for_change_scores: np.ndarray):
        """
        Calculate a proposal distribution over edges such that edges with high influence will get higher probabilities.
        The transformation of total influences into a distribution is done by softmax.
        Parameters
        ----------
        net_for_change_scores : np.ndarray
            The network using which the change scores for all edges is calculated.
        """
        edge_influence = self._calc_edge_influence_on_features(net_for_change_scores)
        self._edge_proposal_dists['features_influence__softmax'] = softmax(edge_influence)

    def sample(self,
               initial_state,
               num_of_nets,
               replace=True,
               edge_proposal_method="uniform",
               # TODO - these two params need to be dependent on the network size
               burn_in=1000,
               steps_per_sample=10,
               node_feature_names={},
               node_features_n_categories={},
               edge_node_flip_ratio=None):
        """
        Sample networks using the Metropolis-Hastings algorithm.
        
        Parameters
        ----------
        initial_state : np.ndarray
            The initial network to start the Markov Chain from
        
        num_of_nets : int
            The number of networks to sample

        burn_in : int
            Optional. The number of burn-in steps for the sampler (number of steps in the chain that are discarded
            before the sampler starts to take samples). *Defaults to 1000*.

        steps_per_sample : int
            Optional. The number of steps to advance the chain between samples. *Defaults to 10*.
        
        replace : bool
            A boolean flag indicating whether we sample with our without replacement. replace=True means networks can be
            duplicated.

        edge_proposal_method : str
            Optional. The method for the MCMC proposal distribution. This is defined as a distribution over the edges
            of the network, which implies how to choose a proposed graph out of all graphs that are 1-edge-away from the
            current graph. *Defaults to "uniform"*.
        """
        current_network = initial_state.copy()

        net_size = current_network.shape[0]
        num_of_node_features = current_network.shape[1] - current_network.shape[0]
        inverse_node_feature_names = {}
        for feature_name, feature_indices in node_feature_names.items():
            for i in feature_indices:
                inverse_node_feature_names[i] = feature_name

        sampled_networks = np.zeros((net_size, net_size + num_of_node_features, num_of_nets))
        sampled_node_features = np.zeros((net_size, num_of_node_features, num_of_nets))
        if edge_node_flip_ratio is None:
            if num_of_node_features > 0:
                edge_node_flip_ratio = max(1, int(net_size / num_of_node_features))
            else:
                edge_node_flip_ratio = net_size

        num_flips = burn_in + (num_of_nets * steps_per_sample)
        if edge_proposal_method == 'uniform':
            edges_to_flip = get_uniform_random_edges_to_flip(net_size, num_flips)
        elif edge_proposal_method == 'features_influence__sum':
            if edge_proposal_method not in self._edge_proposal_dists.keys():
                self._calc_proposal_dist_features_influence__sum(current_network)
            edges_to_flip = get_custom_distribution_random_edges_to_flip(num_flips, self._edge_proposal_dists[
                'features_influence__sum'])
        elif edge_proposal_method == 'features_influence__softmax':
            if edge_proposal_method not in self._edge_proposal_dists.keys():
                self._calc_proposal_dist_features_influence__softmax(current_network)
            edges_to_flip = get_custom_distribution_random_edges_to_flip(num_flips, self._edge_proposal_dists[
                'features_influence__softmax'])
        else:
            raise ValueError(f"Got an unsupported edge proposal method {edge_proposal_method}")

        # for now, nodes_to_flip supports only uniform sampling
        if num_of_node_features != 0:
            nodes_to_flip = get_uniform_random_nodes_to_flip(net_size, num_flips)
            node_features_to_flip = np.random.choice(num_of_node_features, size=num_flips)
            node_features_inds_to_n_categories = {feature_ind: node_features_n_categories[feature_name] for feature_ind, feature_name in inverse_node_feature_names.items()}
            new_node_feature_categories = get_uniform_random_new_node_feature_categories(node_features_to_flip, node_features_inds_to_n_categories)

        random_nums_for_change_acceptance = np.random.rand(num_flips)

        networks_count = 0
        mcmc_iter_count = 0
        feature_flip_counts = {feature_ind: 0 for feature_ind in range(num_of_node_features)}

        t1 = time.time()
        while networks_count != num_of_nets:
            # Edge flip:
            random_edge_entry = edges_to_flip[:, mcmc_iter_count % edges_to_flip.shape[1]] # the number in axis 1 is always just mcmc_iter_count?
            edge_flip_info = {
                'edge': random_edge_entry,
            } # This will be convenient when we'll start using non-binary edges

            # Node flip
            if (num_of_node_features != 0) and (mcmc_iter_count % edge_node_flip_ratio == 0):
                random_node_entry = nodes_to_flip[mcmc_iter_count % nodes_to_flip.shape[0]]
                random_feature_to_flip = node_features_to_flip[mcmc_iter_count % nodes_to_flip.shape[0]]
                current_category = current_network[random_node_entry, random_feature_to_flip]
                # random_feature_name_to_flip = inverse_node_feature_names[random_feature_to_flip]
                random_new_category = new_node_feature_categories[random_feature_to_flip][current_category][feature_flip_counts[random_feature_to_flip]]
                feature_flip_counts[random_feature_to_flip] += 1

            else:
                random_node_entry = None
                random_feature_to_flip = None
                current_category = None
                random_feature_name_to_flip = None
                random_new_category = None
            node_flip_info = {
                'node': random_node_entry,
                'feature': random_feature_to_flip,
                'new_category': random_new_category,
            }

            change_score = self._calculate_weighted_change_score(current_network, edge_flip_info, node_flip_info)

            rand_num = random_nums_for_change_acceptance[mcmc_iter_count % edges_to_flip.shape[1]]
            perform_change = change_score >= 1 or rand_num <= min(1, np.exp(change_score))
            if perform_change:
                self._flip_network_edge(current_network, edge_flip_info['edge'][0], edge_flip_info['edge'][1])
                if random_new_category is not None: # a node flip was suggested
                    self._update_node_features(current_network,
                                               node_flip_info['node'],
                                               node_flip_info['feature'],
                                               node_flip_info['new_category'])

            if (mcmc_iter_count - burn_in) % steps_per_sample == 0:
                sampled_networks[:, :, networks_count] = current_network

                if not replace:
                    # TODO - Since we're flipping coins in the beginning, we're reusing them in case of the non-replacement (since we might need more coin flips that num of networks).
                    #  Consider reflipping coins to get better "pseudorandom properties", in the case that we run out of pre-flipped coins
                    if np.unique(sampled_networks[:, :, :networks_count + 1], axis=2).shape[2] == networks_count + 1:
                        networks_count += 1
                    else:
                        sampled_networks[:, :, networks_count] = np.zeros((net_size, net_size))
                else:
                    networks_count += 1
                    t2 = time.time()
                    if networks_count % 100 == 0:
                        print(f"Sampled {networks_count}/{num_of_nets} networks, time taken: {t2-t1}")

            mcmc_iter_count += 1
        return sampled_networks



import unittest
from pyERGM.utils import *
from pyERGM.metrics import MetricsCollection, NumberOfEdgesDirected, TotalReciprocity
from scipy.linalg import eigh

from matplotlib import pyplot as plt
import sys


class GeneralUtilsTester(unittest.TestCase):
    def test_get_sorted_type_pairs(self):
        types = ["A", "B", "B", "B"]
        sorted_type_pairs = get_sorted_type_pairs(types)
        expected_sorted_type_pairs = [("A", "A"), ("A", "B"), ("B", "A"), ("B", "B")]
        self.assertTrue(sorted_type_pairs == expected_sorted_type_pairs)

    def test_convert_non_flat_non_diag_to_i_j(self):
        n = 4
        i_j = convert_flat_no_diag_idx_to_i_j([7], n)
        expected_i_j = np.array([[2],
                                 [1]])
        self.assertTrue(np.all(expected_i_j == i_j))

        flat_no_diags = np.array([0, 5, 11])
        expected_i_js = np.array([[0, 1, 3],
                                  [1, 3, 2]])
        i_js = convert_flat_no_diag_idx_to_i_j(flat_no_diags, n)
        self.assertTrue(np.all(expected_i_js == i_js))

        with self.assertRaises(IndexError):
            convert_flat_no_diag_idx_to_i_j([13], n)


class TestGreatestConvexMinorant(unittest.TestCase):
    DO_PLOT = False

    def test_not_change_values_of_convex_functions(self):
        xs = np.arange(10) + 1
        convex_func_vals = 1 / xs ** 2

        minorant_vals = get_greatest_convex_minorant(xs, convex_func_vals)

        if TestGreatestConvexMinorant.DO_PLOT:
            plt.plot(xs, convex_func_vals, '.k', label='input values')
            plt.plot(xs, minorant_vals, label='greatest convex minorant')
            plt.legend()
            plt.show()

        self.assertTrue(np.all(minorant_vals == convex_func_vals))

    def test_engineered_scenario(self):
        def get_expected_value(convex_func_vals, wrapping_indices, perturbed_idx):
            slope = (convex_func_vals[wrapping_indices[1]] - convex_func_vals[wrapping_indices[0]]) / (
                    xs[wrapping_indices[1]] - xs[wrapping_indices[0]])
            return slope * xs[perturbed_idx] + convex_func_vals[wrapping_indices[1]] - xs[wrapping_indices[1]] * slope

        num_points = 10
        xs = np.arange(num_points) + 1
        convex_func_vals = np.exp(-xs)
        num_scenarios = 4
        perturbation = 0.5
        indices_to_increase = np.array([[1, 3, 7], [1, 2, 3], [0, 2, 4], [5, 6, 9]]).astype(int)
        expected_values = np.array([convex_func_vals, ] * num_scenarios)
        expected_values[0, 1] = get_expected_value(convex_func_vals, (0, 2), 1)
        expected_values[0, 3] = get_expected_value(convex_func_vals, (2, 4), 3)
        expected_values[0, 7] = get_expected_value(convex_func_vals, (6, 8), 7)

        expected_values[1, 1] = get_expected_value(convex_func_vals, (0, 4), 1)
        expected_values[1, 2] = get_expected_value(convex_func_vals, (0, 4), 2)
        expected_values[1, 3] = get_expected_value(convex_func_vals, (0, 4), 3)

        expected_values[2, 0] = convex_func_vals[0] + perturbation
        expected_values[2, 2] = get_expected_value(convex_func_vals, (1, 3), 2)
        expected_values[2, 4] = get_expected_value(convex_func_vals, (3, 5), 4)

        expected_values[3, 5] = get_expected_value(convex_func_vals, (4, 7), 5)
        expected_values[3, 6] = get_expected_value(convex_func_vals, (4, 7), 6)
        expected_values[3, 9] = convex_func_vals[9] + perturbation

        for i in range(num_scenarios):
            cur_vals = convex_func_vals.copy()
            cur_vals[indices_to_increase[i]] += perturbation
            minorant_vals = get_greatest_convex_minorant(xs, cur_vals)
            if TestGreatestConvexMinorant.DO_PLOT:
                plt.plot(xs, cur_vals, '.k', label='input values')
                plt.plot(xs, minorant_vals, label='greatest convex minorant')
                plt.legend()
                plt.title(f'scenario {i}')
                plt.show()

            for j in range(num_points):
                self.assertAlmostEqual(minorant_vals[j], expected_values[i, j], places=8)

    def test_random_points(self):
        np.random.seed(8972634)
        num_points = 100
        xs = np.arange(num_points) + 1
        values = 10 * np.random.rand(100)
        minorant_vals = get_greatest_convex_minorant(xs, values)
        if TestGreatestConvexMinorant.DO_PLOT:
            plt.plot(xs, values, '.k', label='input values')
            plt.plot(xs, minorant_vals, label='greatest convex minorant')
            plt.legend()
            plt.show()
        self.assertTrue(np.all(minorant_vals <= values))
        self.assertTrue(np.all(np.diff(minorant_vals, n=2) >= -10 ** -10))


class TestSparseTensorUtilities(unittest.TestCase):
    def test_transpose_sample_matrices(self):
        np_tensor = np.random.choice([0, 1], size=(5, 5, 1), p=[0.8, 0.2])
        expected_np_tensor_T = np.transpose(np_tensor, axes=(1, 0, 2))

        sparse_tensor = np_tensor_to_sparse_tensor(np_tensor)

        self.assertTrue(sparse_tensor.is_sparse)

        transposed_sparse_tensor = transpose_sparse_sample_matrices(sparse_tensor)
        transposed_sparse_tensor_as_np = transposed_sparse_tensor.to_dense().numpy()

        self.assertTrue(np.all(expected_np_tensor_T == transposed_sparse_tensor_as_np))


class TestCovarianceMatrixEstimation(unittest.TestCase):

    @staticmethod
    def validate_test_sample_validity(metrics_collection, sample):
        features_of_sample = metrics_collection.calculate_sample_statistics(sample)
        gamma_hats = approximate_auto_correlation_function(features_of_sample)
        capital_gammas = calc_capital_gammas(gamma_hats)
        Sigma_0 = -gamma_hats[0] + 2 * capital_gammas[0]
        Sigma_1 = Sigma_0 + 2 * capital_gammas[1]
        Sigma_2 = Sigma_1 + 2 * capital_gammas[2]

        is_Sigma_0_positive_definite = eigh(Sigma_0)[0].min() > 0
        is_Sigma_1_positive_definite = eigh(Sigma_1)[0].min() > 0
        is_Sigma_2_positive_definite = eigh(Sigma_2)[0].min() > 0

        det_Sigma_0 = np.linalg.det(Sigma_0)
        det_Sigma_1 = np.linalg.det(Sigma_1)
        det_Sigma_2 = np.linalg.det(Sigma_2)

        if ((is_Sigma_0_positive_definite and det_Sigma_0 < det_Sigma_1) or
                (is_Sigma_1_positive_definite and det_Sigma_1 < det_Sigma_2)):
            return True
        return False

    @staticmethod
    def find_sample_for_multivariate_initial_seq_covariance_matrix_estimation_test(num_trials=1000, sample_size=6, n=3,
                                                                                   p=0.5):
        metrics_collection = MetricsCollection([NumberOfEdgesDirected(), TotalReciprocity()], True, n)
        for i in range(num_trials):
            sample = np.zeros((n, n, sample_size))
            for j in range(sample_size):
                cur_mat = nx.to_numpy_array(nx.erdos_renyi_graph(n, p, directed=True))
                sample[:, :, j] = cur_mat
                if TestCovarianceMatrixEstimation.validate_test_sample_validity(metrics_collection, sample):
                    return sample
        print("No sample was valid")
        return None

    @staticmethod
    def _construct_sample():
        n = 3
        sample_size = 6
        W1 = np.array([[0, 1, 0],
                       [0, 0, 1],
                       [1, 1, 0]])

        W2 = np.array([[0, 1, 0],
                       [1, 0, 0],
                       [0, 1, 0]])

        W3 = np.array([[0, 0, 1],
                       [0, 0, 1],
                       [1, 1, 0]])

        W4 = np.array([[0, 0, 1],
                       [1, 0, 1],
                       [1, 1, 0]])

        W5 = np.array([[0, 0, 1],
                       [1, 0, 0],
                       [1, 0, 0]])

        W6 = np.array([[0, 0, 1],
                       [1, 0, 1],
                       [1, 0, 0]])

        sample = np.zeros((n, n, sample_size))
        sample[:, :, 0] = W1
        sample[:, :, 1] = W2
        sample[:, :, 2] = W3
        sample[:, :, 3] = W4
        sample[:, :, 4] = W5
        sample[:, :, 5] = W6

        return sample

    def test_batches_covariance_estimation(self):
        sample = self._construct_sample()
        n = sample.shape[0]

        expected_covariance_batch_estimation = np.ones((2, 2)) * 4 / 9

        network_statistics = MetricsCollection([NumberOfEdgesDirected(), TotalReciprocity()], True, n)
        features_of_sample = network_statistics.calculate_sample_statistics(sample)
        mean_features = features_of_sample.mean(axis=1)

        sys.setrecursionlimit(2000)
        batch_estimation = covariance_matrix_estimation(features_of_sample, mean_features, method='batch',
                                                        num_batches=3)
        self.assertTrue(np.abs(expected_covariance_batch_estimation - batch_estimation).max() < 10 ** -15)

    def test_naive_covariance_matrix_estimation(self):
        sample = self._construct_sample()
        n = sample.shape[0]
        expected_covariance_naive_estimation = np.array([[17 / 36, 8 / 36], [8 / 36, 8 / 36]])

        network_statistics = MetricsCollection([NumberOfEdgesDirected(), TotalReciprocity()], True, n)
        features_of_sample = network_statistics.calculate_sample_statistics(sample)
        mean_features = features_of_sample.mean(axis=1)

        sys.setrecursionlimit(2000)
        naive_estimation = covariance_matrix_estimation(features_of_sample, mean_features, method='naive')
        self.assertTrue(np.abs(expected_covariance_naive_estimation - naive_estimation).max() < 10 ** -14)

    def test_auto_correlation_function(self):
        n = 3
        sample_size = 3
        W1 = np.array([[0, 1, 0],
                       [0, 0, 1],
                       [1, 1, 0]])

        W2 = np.array([[0, 1, 0],
                       [0, 0, 1],
                       [0, 1, 0]])

        W3 = np.array([[0, 1, 1],
                       [0, 0, 1],
                       [1, 1, 0]])

        sample = np.zeros((n, n, sample_size))
        sample[:, :, 0] = W1
        sample[:, :, 1] = W2
        sample[:, :, 2] = W3

        expected_gamma_hat_0 = 1 / 9 * np.array([[6, 3], [3, 2]])
        expected_gamma_hat_1 = 1 / 27 * np.array([[-9, -6], [0, -1]])
        expected_gamma_hat_2 = 1 / 27 * np.array([[0, 0], [-3, -2]])

        network_statistics = MetricsCollection([NumberOfEdgesDirected(), TotalReciprocity()], True, n)

        features_of_sample = network_statistics.calculate_sample_statistics(sample)
        gammas = approximate_auto_correlation_function(features_of_sample)

        self.assertTrue(np.abs(expected_gamma_hat_0 - gammas[0]).max() < 10 ** -15)
        self.assertTrue(np.abs(expected_gamma_hat_1 - gammas[1]).max() < 10 ** -15)
        self.assertTrue(np.abs(expected_gamma_hat_2 - gammas[2]).max() < 10 ** -15)

    def test_capital_gammas(self):
        n = 3
        sample_size = 4

        W1 = np.array([[0, 0, 1],
                       [0, 0, 1],
                       [1, 1, 0]])

        W2 = np.array([[0, 0, 1],
                       [0, 0, 1],
                       [1, 1, 0]])

        W3 = np.array([[0, 1, 0],
                       [1, 0, 0],
                       [0, 1, 0]])

        W4 = np.array([[0, 0, 1],
                       [1, 0, 0],
                       [1, 0, 0]])

        sample = np.zeros((n, n, sample_size))
        sample[:, :, 0] = W1
        sample[:, :, 1] = W2
        sample[:, :, 2] = W3
        sample[:, :, 3] = W4

        expected_capital_gamma_0 = 5 / 16 * np.ones((2, 2))
        expected_capital_gamma_1 = -3 / 16 * np.ones((2, 2))

        network_statistics = MetricsCollection([NumberOfEdgesDirected(), TotalReciprocity()], True, n)

        features_of_sample = network_statistics.calculate_sample_statistics(sample)
        gammas = approximate_auto_correlation_function(features_of_sample)
        capital_gammas = calc_capital_gammas(gammas)

        self.assertTrue(np.abs(capital_gammas[0] - expected_capital_gamma_0).max() < 10 ** -15)
        self.assertTrue(np.abs(capital_gammas[1] - expected_capital_gamma_1).max() < 10 ** -15)

    def test_single_auto_correlation_function(self):
        n = 3
        sample_size = 3
        W1 = np.array([[0, 1, 0],
                       [0, 0, 1],
                       [1, 1, 0]])

        W2 = np.array([[0, 1, 0],
                       [0, 0, 1],
                       [0, 1, 0]])

        W3 = np.array([[0, 1, 1],
                       [0, 0, 1],
                       [1, 1, 0]])

        sample = np.zeros((n, n, sample_size))
        sample[:, :, 0] = W1
        sample[:, :, 1] = W2
        sample[:, :, 2] = W3

        expected_gamma_hat_0 = 1 / 9 * np.array([[6, 3], [3, 2]])
        expected_gamma_hat_1 = 1 / 27 * np.array([[-9, -6], [0, -1]])
        expected_gamma_hat_2 = 1 / 27 * np.array([[0, 0], [-3, -2]])
        expected_gammas = [expected_gamma_hat_0, expected_gamma_hat_1, expected_gamma_hat_2]

        network_statistics = MetricsCollection([NumberOfEdgesDirected(), TotalReciprocity()], True, n)

        features_of_sample = network_statistics.calculate_sample_statistics(sample)
        features_of_sample_mean = features_of_sample.mean(axis=1)
        diff_from_mean = features_of_sample - features_of_sample_mean[:, None]
        for k in range(3):
            cur_gamma = approximate_kth_auto_correlation_function(diff_from_mean, k)
            self.assertTrue(np.abs(expected_gammas[k] - cur_gamma).max() < 10 ** -15)

    def test_kth_capital_gamma(self):
        n = 3
        sample_size = 4

        W1 = np.array([[0, 0, 1],
                       [0, 0, 1],
                       [1, 1, 0]])

        W2 = np.array([[0, 0, 1],
                       [0, 0, 1],
                       [1, 1, 0]])

        W3 = np.array([[0, 1, 0],
                       [1, 0, 0],
                       [0, 1, 0]])

        W4 = np.array([[0, 0, 1],
                       [1, 0, 0],
                       [1, 0, 0]])

        sample = np.zeros((n, n, sample_size))
        sample[:, :, 0] = W1
        sample[:, :, 1] = W2
        sample[:, :, 2] = W3
        sample[:, :, 3] = W4

        expected_capital_gamma_0 = 5 / 16 * np.ones((2, 2))
        expected_capital_gamma_1 = -3 / 16 * np.ones((2, 2))

        network_statistics = MetricsCollection([NumberOfEdgesDirected(), TotalReciprocity()], True, n)

        features_of_sample = network_statistics.calculate_sample_statistics(sample)
        features_of_sample_mean = features_of_sample.mean(axis=1)
        diff_from_mean = features_of_sample - features_of_sample_mean[:, None]
        capital_gamma_0 = calc_kth_capital_gamma(diff_from_mean, 0)
        capital_gamma_1 = calc_kth_capital_gamma(diff_from_mean, 1)
        self.assertTrue(np.abs(capital_gamma_0 - expected_capital_gamma_0).max() < 10 ** -15)
        self.assertTrue(np.abs(capital_gamma_1 - expected_capital_gamma_1).max() < 10 ** -15)

    def test_multivariate_initial_seq_covariance_matrix_estimation(self):
        n = 3
        sample_size = 6
        W1 = np.array([[0, 0, 0],
                       [0, 0, 1],
                       [1, 0, 0]])
        W2 = np.array([[0, 1, 0],
                       [0, 0, 0],
                       [1, 1, 0]])
        W3 = np.array([[0, 1, 1],
                       [0, 0, 1],
                       [1, 0, 0]])
        W4 = np.array([[0, 1, 0],
                       [0, 0, 0],
                       [0, 1, 0]])
        W5 = np.array([[0, 1, 0],
                       [0, 0, 1],
                       [1, 1, 0]])
        W6 = np.array([[0, 1, 1],
                       [1, 0, 0],
                       [1, 1, 0]])
        sample = np.zeros((n, n, sample_size))
        sample[:, :, 0] = W1
        sample[:, :, 1] = W2
        sample[:, :, 2] = W3
        sample[:, :, 3] = W4
        sample[:, :, 4] = W5
        sample[:, :, 5] = W6

        # NOTE! this matrix was calculated after automatically calculating the capital gammas (using the dedicated
        # function that has a dedicated test). I gave up on calculating everything from scratch after a long battle.
        # Maybe we should come to that in the future.
        expected_covariance_multivariate_initial_seq_estimation = 1 / (6 * 36) * np.array([[264, 192], [192, 144]])

        network_statistics = MetricsCollection([NumberOfEdgesDirected(), TotalReciprocity()], True, n)
        features_of_sample = network_statistics.calculate_sample_statistics(sample)
        mean_features = features_of_sample.mean(axis=1)

        sys.setrecursionlimit(2000)
        multivariate_initial_seq_estimation = covariance_matrix_estimation(features_of_sample, mean_features,
                                                                           method='multivariate_initial_sequence')
        self.assertTrue(np.abs(
            expected_covariance_multivariate_initial_seq_estimation - multivariate_initial_seq_estimation).max() <
                        10 ** -14)



import shutil
from abc import ABC, abstractmethod
from typing import Collection, Callable
from copy import deepcopy

import numpy as np
import pandas as pd
from scipy.spatial.distance import pdist, squareform

from pyERGM.utils import *


class Metric(ABC):
    def __init__(self, requires_graph=False, metric_type='binary_edge', metric_node_feature=None):
        self.requires_graph = requires_graph
        # Each metric either expects directed or undirected graphs. This field should be initialized in the constructor
        # and should not change.
        self._is_directed = None
        self._is_dyadic_independent = True
        self._n_nodes = None
        self._indices_to_ignore = None
        self._metric_type = metric_type # can have values "node", "binary_edge", "non_binary_edge"
        if self._metric_type not in ['node', 'binary_edge', 'binary_edge']:
            raise ValueError(f"invalid metric type: {self._metric_type}. Should be one of: 'node', 'binary_edge', 'binary_edge'")

        self.metric_node_feature = metric_node_feature # relevant only if metric_type='node'

    def initialize_indices_to_ignore(self):
        self._indices_to_ignore = np.array([False] * self._get_total_feature_count())

        if hasattr(self, "_indices_from_user") and self._indices_from_user is not None:
            self.update_indices_to_ignore(self._indices_from_user)

    def _handle_indices_to_ignore(self, res, axis=0):
        if self._indices_to_ignore is None:
            return res

        if axis > 1:
            raise ValueError("Axis should be 0 or 1")
        if axis == 1:
            return res[:, ~self._indices_to_ignore]
        return res[~self._indices_to_ignore]

    @abstractmethod
    def calculate(self, input: np.ndarray | nx.Graph):
        pass

    def _get_effective_feature_count(self):
        """
        How many features does this metric produce. Defaults to 1.
        """
        if self._indices_to_ignore is None:
            return self._get_total_feature_count()
        return np.sum(~self._indices_to_ignore)

    def _get_total_feature_count(self):
        """
        How many features does this metric produce, including the ignored ones. Defaults to 1
        """
        return 1

    def update_indices_to_ignore(self, indices_to_ignore):
        self._indices_to_ignore[indices_to_ignore] = True

    def calc_change_score(self, current_network: np.ndarray | nx.Graph, indices: tuple):
        """
        The default naive way to calculate the change score (namely, the difference in statistics) of a pair of
        networks.

        The newly proposed network is created by flipping the edge denoted by `indices`

        Returns
        -------
        statistic of proposed_network minus statistic of current_network.
        """
        i, j = indices
        if self.requires_graph:
            proposed_network = current_network.copy()
            if proposed_network.has_edge(i, j):
                proposed_network.remove_edge(i, j)
            else:
                proposed_network.add_edge(i, j)
        else:
            proposed_network = current_network.copy()
            proposed_network[i, j] = 1 - proposed_network[i, j]

            if not self._is_directed:
                proposed_network[j, i] = 1 - proposed_network[j, i]

        proposed_network_stat = self.calculate(proposed_network)
        current_network_stat = self.calculate(current_network)
        return proposed_network_stat - current_network_stat

    def calculate_for_sample(self, networks_sample: np.ndarray | Collection[nx.Graph]):
        num_of_samples = networks_sample.shape[2]

        result = np.zeros((self._get_effective_feature_count(), num_of_samples))
        for i in range(num_of_samples):
            network = networks_sample[i] if self.requires_graph else networks_sample[:, :, i]
            result[:, i] = self.calculate(network)
        return result

    def calculate_mple_regressors(self, observed_network: np.ndarray | nx.Graph, edges_indices_lims: tuple[int]):
        num_edges_to_take = edges_indices_lims[1] - edges_indices_lims[0]
        Xs = np.zeros((num_edges_to_take, self._get_effective_feature_count()))
        edge_idx = 0

        if self._is_directed:
            for i in range(self._n_nodes):
                for j in range(self._n_nodes):
                    if i == j:
                        continue
                    if edge_idx < edges_indices_lims[0]:
                        edge_idx += 1
                        continue
                    indices = (i, j)
                    observed_edge_off = observed_network.copy()
                    if self.requires_graph:
                        if observed_edge_off.has_edge(i, j):
                            observed_edge_off.remove_edge(i, j)
                    else:
                        observed_edge_off[i, j] = 0

                    Xs[edge_idx - edges_indices_lims[0]] = self.calc_change_score(observed_edge_off, indices)
                    edge_idx += 1
                    if edge_idx == edges_indices_lims[1]:
                        return Xs

        else:
            Xs = np.zeros((num_edges_to_take, self._get_effective_feature_count()))
            for i in range(self._n_nodes - 1):
                for j in range(i + 1, self._n_nodes):
                    if edge_idx < edges_indices_lims[0]:
                        edge_idx += 1
                        continue

                    indices = (i, j)

                    observed_edge_off = observed_network.copy()
                    if self.requires_graph:
                        if observed_edge_off.has_edge(i, j):
                            observed_edge_off.remove_edge(i, j)
                    else:
                        observed_edge_off[i, j] = 0
                        observed_edge_off[j, i] = 0

                    Xs[edge_idx - edges_indices_lims[0]] = self.calc_change_score(observed_edge_off, indices)
                    edge_idx += 1
                    if edge_idx == edges_indices_lims[1]:
                        return Xs

    def _get_metric_names(self):
        """
        Get the names of the features that this metric produces. Defaults to the name of the metric if the metric creates a single feature, and multiple
        names if the metric creates multiple features.

        Returns
        -------
        parameter_names: tuple
            A tuple of strings, each string is the name of a parameter that this metric produces.
        """
        total_n_features = self._get_total_feature_count()
        if total_n_features == 1:
            return (str(self),)
        else:
            parameter_names = ()
            for i in range(total_n_features):
                if self._indices_to_ignore is not None and self._indices_to_ignore[i]:
                    continue
                parameter_names += (f"{str(self)}_{i + 1}",)
            return parameter_names

    def _get_ignored_features(self):
        if self._indices_to_ignore is None or not np.any(self._indices_to_ignore):
            return tuple()

        ignored_features = ()
        for i in range(self._get_total_feature_count()):
            if self._indices_to_ignore is not None and self._indices_to_ignore[i]:
                ignored_features += (f"{str(self)}_{i + 1}",)

        return ignored_features

    def _calc_bootstrapped_scalar_feature(self, first_halves_to_use: np.ndarray,
                                          first_halves_indices: np.ndarray[int], second_halves_indices: np.ndarray[int],
                                          max_feature_val_calculator: Callable):
        num_nodes_in_observed = first_halves_indices.shape[0] + second_halves_indices.shape[0]
        num_nodes_in_first_half = first_halves_indices.shape[0]
        max_feature_in_observed = max_feature_val_calculator(num_nodes_in_observed)
        max_feature_in_first_half = max_feature_val_calculator(num_nodes_in_first_half)
        return self.calculate_for_sample(
            first_halves_to_use) * max_feature_in_observed / max_feature_in_first_half


class NumberOfEdgesUndirected(Metric):
    def __str__(self):
        return "num_edges_undirected"

    def __init__(self):
        super().__init__(requires_graph=False)
        self._is_directed = False
        self._is_dyadic_independent = True

    def calculate(self, W: np.ndarray):
        return np.sum(W) // 2

    def calc_change_score(self, current_network: np.ndarray, indices: tuple):
        return -1 if current_network[indices[0], indices[1]] else 1

    def calculate_for_sample(self, networks_sample: np.ndarray | torch.Tensor):
        """
        Sum each matrix over all matrices in sample
        """
        return networks_sample.sum(axis=(0, 1)) // 2

    def calculate_bootstrapped_features(self, first_halves_to_use: np.ndarray,
                                        second_halves_to_use: np.ndarray,
                                        first_halves_indices: np.ndarray[int], second_halves_indices: np.ndarray[int]):
        """
        Calculates the bootstrapped number of edges, by counting edges in the sampled subnetworks, and normalizing by
        network size (i.e., calculating the fraction of existing edges out of all possible ones in sampled subnetworks,
        and multiplying by the number of possible edges in the full observed network).
        Parameters
        ----------
        first_halves_to_use
            Multiple samples of subnetworks of an observed network, representing the connectivity between half of the
            nodes in the large network.
        second_halves_to_use
            The subnetworks formed by the complementary set of nodes of the large network for each sample.
        first_halves_indices
            The indices of the nodes in the first half of the large network for each sample, according to the ordering
            of the nodes in the large network.
        second_halves_indices
            The indices of the nodes in the second half of the large network for each sample, according to the ordering

        Returns
        -------
        Properly normalized statistics of subnetworks of an observed network.
        """
        return self._calc_bootstrapped_scalar_feature(first_halves_to_use, first_halves_indices, second_halves_indices,
                                                      lambda n: n * (n - 1) / 2)


class NumberOfEdgesDirected(Metric):
    def __str__(self):
        return "num_edges_directed"

    def __init__(self):
        super().__init__(requires_graph=False)
        self._is_directed = True
        self._is_dyadic_independent = True

    def calculate(self, W: np.ndarray):
        return np.sum(W)

    @staticmethod
    @njit
    def calc_change_score(current_network: np.ndarray, indices: tuple):
        return -1 if current_network[indices[0], indices[1]] else 1

    @staticmethod
    @calc_for_sample_njit()
    def calculate_for_sample(networks_sample: np.ndarray | torch.Tensor):
        """
        Sum each matrix over all matrices in sample
        """
        return networks_sample.sum(axis=0).sum(axis=0)

    @staticmethod
    # TODO: understand how to properly use numba here (the dtype=bool makes some trouble).
    # @njit
    def calculate_change_score_full_network(current_network: np.ndarray):
        sign_matrix = np.where(current_network == 1, -1,
                               1)  # -1 for edges that exist (because we want to remove them), 1 for edges that don't exist (because we want to add them).
        nondiag_signs = sign_matrix[~np.eye(sign_matrix.shape[0], dtype=bool)].flatten()

        return nondiag_signs[:, np.newaxis]

    @staticmethod
    @njit
    def calculate_mple_regressors(observed_network: np.ndarray, edges_indices_lims: tuple[int]):
        return np.ones((edges_indices_lims[1] - edges_indices_lims[0], 1))

    def calculate_bootstrapped_features(self, first_halves_to_use: np.ndarray,
                                        second_halves_to_use: np.ndarray,
                                        first_halves_indices: np.ndarray[int], second_halves_indices: np.ndarray[int]):
        """
        Calculates the bootstrapped number of edges, by counting edges in the sampled subnetworks, and normalizing by
        network size (i.e., calculating the fraction of existing edges out of all possible ones in sampled subnetworks,
        and multiplying by the number of possible edges in the full observed network).
        Parameters
        ----------
        first_halves_to_use
            Multiple samples of subnetworks of an observed network, representing the connectivity between half of the
            nodes in the large network.
        second_halves_to_use
            The subnetworks formed by the complementary set of nodes of the large network for each sample.
        first_halves_indices
            The indices of the nodes in the first half of the large network for each sample, according to the ordering
            of the nodes in the large network.
        second_halves_indices
            The indices of the nodes in the second half of the large network for each sample, according to the ordering

        Returns
        -------
        Properly normalized statistics of subnetworks of an observed network.
        """
        return self._calc_bootstrapped_scalar_feature(first_halves_to_use, first_halves_indices, second_halves_indices,
                                                      lambda n: n * (n - 1))


# TODO: change the name of this one to undirected and implement also a directed version?
class NumberOfTriangles(Metric):
    def __str__(self):
        return "num_triangles"

    def __init__(self):
        super().__init__(requires_graph=False)
        self._is_directed = False
        self._is_dyadic_independent = False

    def calculate(self, W: np.ndarray):
        if not np.all(W.T == W):
            raise ValueError("NumOfTriangles not implemented for directed graphs")
        # the (i,j)-th entry of W^3 counts the number of 3-length paths from node i to node j. Thus, the i-th element on
        # the diagonal counts the number of triangles that node 1 is part of (3-length paths from i to itself). As the
        # graph is undirected, we get that each path is counted twice ("forward and backwards"), thus the division by 2.
        # Additionally, each triangle is counted 3 times by diagonal elements (once for each node that takes part in
        # forming it), thus the division by 3.
        return (np.linalg.matrix_power(W, 3)).diagonal().sum() // (3 * 2)

    def calc_change_score(self, current_network: np.ndarray, indices: tuple):
        # The triangles that are affected by the edge toggling are those that involve it, namely, if the (i,j)-th edge
        # is toggled, the change in absolute value equals to the number of nodes k for which the edges (i,k) and (j,k)
        # exist. This is equivalent to the number of 2-length paths from i to j, which is the (i,j)-th entry of W^2.
        # If the edge is turned on, the change is positive, and otherwise negative.

        sign = -1 if current_network[indices[0], indices[1]] else 1
        return sign * np.dot(current_network[indices[0]], current_network[:, indices[1]])

    def calculate_for_sample(self, networks_sample: np.ndarray):
        return np.einsum('ijk,jlk,lik->k', networks_sample, networks_sample, networks_sample) // (3 * 2)


class BaseDegreeVector(Metric):
    """
    A base class for calculating a degree vector for a network.
    To avoid multicollinearity with other features, an optional parameter `indices_to_ignore` can be used to specify
    which indices the calculation ignores.
    """

    def __init__(self, is_directed: bool, indices_from_user=None):
        super().__init__(requires_graph=False)

        self._indices_from_user = indices_from_user.copy() if indices_from_user is not None else None

        self._is_directed = is_directed

    def _get_total_feature_count(self):
        return self._n_nodes

    def calculate_bootstrapped_features(self, first_halves_to_use: np.ndarray,
                                        second_halves_to_use: np.ndarray,
                                        first_halves_indices: np.ndarray[int], second_halves_indices: np.ndarray[int]):
        """
        Calculates the bootstrapped degree, by counting the connections of each node in the sampled subnetworks, and
        normalizing by network size (i.e., calculating the fraction of existing edges out of all possible ones for each
        node in sampled subnetworks, and multiplying by the number of possible edges of a single node in the full
        observed network).
        Each node appears in one of the sub-samples, so both are used, and the indices are used to order the calculated
        values in the entire features vector.
        Parameters
        ----------
        first_halves_to_use
            Multiple samples of subnetworks of an observed network, representing the connectivity between half of the
            nodes in the large network.
        second_halves_to_use
            The subnetworks formed by the complementary set of nodes of the large network for each sample.
        first_halves_indices
            The indices of the nodes in the first half of the large network for each sample, according to the ordering
            of the nodes in the large network.
        second_halves_indices
            The indices of the nodes in the second half of the large network for each sample, according to the ordering

        Returns
        -------
        Properly normalized statistics of subnetworks of an observed network.
        """
        num_nodes_in_observed = first_halves_indices.shape[0] + second_halves_indices.shape[0]
        num_nodes_in_first_half = first_halves_indices.shape[0]
        num_nodes_in_second_half = second_halves_indices.shape[0]

        # NOTE! we want the estimated covariance matrix from the bootstrap to match the dimension of samples from the
        # model during optimization, so we must first calculate the degrees for all the nodes (thus the initialization
        # of a new instance, which makes sure that no indices are ignored), and then ignore the indices that will be
        # removed by the metric.
        fresh_instance = self.__class__()  # No child class has a `is_directed` input to the constructor.
        degrees_first_halves = fresh_instance.calculate_for_sample(first_halves_to_use) * (
                num_nodes_in_observed - 1) / (num_nodes_in_first_half - 1)
        fresh_instance._n_nodes = num_nodes_in_second_half
        degrees_second_halves = fresh_instance.calculate_for_sample(second_halves_to_use) * (
                num_nodes_in_observed - 1) / (num_nodes_in_second_half - 1)
        num_sub_samples = first_halves_to_use.shape[2]
        bootstrapped_degrees = np.zeros((self._n_nodes, num_sub_samples))
        bootstrapped_degrees[first_halves_indices, np.arange(num_sub_samples)] = degrees_first_halves
        bootstrapped_degrees[second_halves_indices, np.arange(num_sub_samples)] = degrees_second_halves
        return self._handle_indices_to_ignore(bootstrapped_degrees)


class InDegree(BaseDegreeVector):
    """
    Calculate the in-degree of each node in a directed graph.
    """

    def __str__(self):
        return "indegree"

    def __init__(self, indices_from_user=None):
        super().__init__(is_directed=True, indices_from_user=indices_from_user)
        self._is_dyadic_independent = True

    def calculate(self, W: np.ndarray):
        return self._handle_indices_to_ignore(W.sum(axis=0))

    def calc_change_score(self, current_network: np.ndarray, indices: tuple):
        n = current_network.shape[0]
        diff = np.zeros(n)
        i, j = indices

        sign = -1 if current_network[i, j] else 1

        diff[j] = sign
        return self._handle_indices_to_ignore(diff)

    def calculate_for_sample(self, networks_sample: np.ndarray | torch.Tensor):
        summed_tensor = networks_sample.sum(axis=0)

        if isinstance(networks_sample, torch.Tensor) and networks_sample.is_sparse:
            n_nodes = networks_sample.shape[0]
            n_samples = networks_sample.shape[2]

            indices = self._handle_indices_to_ignore(summed_tensor.indices(), axis=1)
            values = self._handle_indices_to_ignore(summed_tensor.values())
            return torch.sparse_coo_tensor(indices, values, (n_nodes, n_samples))
        else:
            return self._handle_indices_to_ignore(summed_tensor)


class OutDegree(BaseDegreeVector):
    """
    Calculate the out-degree of each node in a directed graph.
    """

    def __str__(self):
        return "outdegree"

    def __init__(self, indices_from_user=None):
        super().__init__(is_directed=True, indices_from_user=indices_from_user)
        self._is_dyadic_independent = True

    def calculate(self, W: np.ndarray):
        return self._handle_indices_to_ignore(W.sum(axis=1))

    def calc_change_score(self, current_network: np.ndarray, indices: tuple):
        n = current_network.shape[0]
        diff = np.zeros(n)
        i, j = indices

        sign = -1 if current_network[i, j] else 1

        diff[i] = sign
        return self._handle_indices_to_ignore(diff)

    def calculate_for_sample(self, networks_sample: np.ndarray | torch.Tensor):
        summed_tensor = networks_sample.sum(axis=1)

        if isinstance(networks_sample, torch.Tensor) and networks_sample.is_sparse:
            n_nodes = networks_sample.shape[0]
            n_samples = networks_sample.shape[2]

            indices = self._handle_indices_to_ignore(summed_tensor.indices(), axis=1)
            values = self._handle_indices_to_ignore(summed_tensor.values())
            return torch.sparse_coo_tensor(indices, values, (n_nodes, n_samples))
        else:
            return self._handle_indices_to_ignore(summed_tensor)


class UndirectedDegree(BaseDegreeVector):
    """
    Calculate the degree of each node in an undirected graph.
    """

    def __str__(self):
        return "undirected_degree"

    def __init__(self, indices_from_user=None):
        super().__init__(is_directed=False, indices_from_user=indices_from_user)
        self._is_dyadic_independent = True

    def calculate(self, W: np.ndarray):
        return self._handle_indices_to_ignore(W.sum(axis=0))

    def calculate_for_sample(self, networks_sample: np.ndarray):
        return self._handle_indices_to_ignore(networks_sample.sum(axis=0))


class Reciprocity(Metric):
    """
    The Reciprocity metric takes the connectivity matrix of a directed graph, and returns a vector
    of size n-choose-2 indicating whether nodes i,j are connected. i.e. $ y_{i, j} \cdot y_{j, i} $
    for every possible pair of nodes   
    """

    def __str__(self):
        return "reciprocity"

    def __init__(self):
        super().__init__(requires_graph=False)
        self._is_directed = True
        self._is_dyadic_independent = False

    def calculate(self, W: np.ndarray):
        return (W * W.T)[np.triu_indices(W.shape[0], 1)]

    def _get_total_feature_count(self):
        # n choose 2
        return self._n_nodes * (self._n_nodes - 1) // 2

    def calc_change_score(self, current_network: np.ndarray, indices: tuple):
        # Note: we intentionally initialize the whole matrix and return np.triu_indices() by the end (rather than
        # initializing an array of zeros of size n choose 2) to ensure compliance with the indexing returned by
        # the calculate method.
        i, j = indices
        all_changes = np.zeros(current_network.shape)
        min_idx = min(indices)
        max_idx = max(indices)

        if current_network[j, i] and not current_network[i, j]:
            all_changes[min_idx, max_idx] = 1
        elif current_network[j, i] and current_network[i, j]:
            all_changes[min_idx, max_idx] = -1
        return all_changes[np.triu_indices(all_changes.shape[0], 1)]


class TotalReciprocity(Metric):
    """
    Calculates how many reciprocal connections exist in a network  
    """

    def __str__(self):
        return "total_reciprocity"

    def __init__(self):
        super().__init__(requires_graph=False)
        self._is_directed = True
        self._is_dyadic_independent = False

    def calculate(self, W: np.ndarray):
        return (W * W.T).sum() / 2

    @staticmethod
    @njit
    def calc_change_score(current_network: np.ndarray, indices: tuple):
        i, j = indices

        if current_network[j, i] and not current_network[i, j]:
            return 1
        elif current_network[j, i] and current_network[i, j]:
            return -1
        else:
            return 0

    @staticmethod
    # @njit # Not supporting neither np.einsum nor sparse torch
    def calculate_for_sample(networks_sample: np.ndarray | torch.Tensor):
        if isinstance(networks_sample, torch.Tensor) and networks_sample.is_sparse:
            transposed_sparse_tensor = transpose_sparse_sample_matrices(networks_sample)
            return torch.sum(networks_sample * transposed_sparse_tensor, axis=(0, 1)) / 2
        elif isinstance(networks_sample, np.ndarray):
            return np.einsum("ijk,jik->k", networks_sample, networks_sample) / 2
        else:
            raise ValueError(f"Unsupported type of sample: {type(networks_sample)}! Supported types are np.ndarray and "
                             f"torch.Tensor with is_sparse=True")

    def calculate_bootstrapped_features(self, first_halves_to_use: np.ndarray,
                                        second_halves_to_use: np.ndarray,
                                        first_halves_indices: np.ndarray[int], second_halves_indices: np.ndarray[int]):
        """
        Calculates the bootstrapped number of reciprocal dyads, by counting such pairs in the sampled subnetworks, and
        normalizing by network size (i.e., calculating the fraction of existing reciprocal dyads out of all possible
        ones in sampled subnetworks, and multiplying by the number of possible reciprocal dyads in the full observed
        network).
        Parameters
        ----------
        first_halves_to_use
            Multiple samples of subnetworks of an observed network, representing the connectivity between half of the
            nodes in the large network.
        second_halves_to_use
            The subnetworks formed by the complementary set of nodes of the large network for each sample.
        first_halves_indices
            The indices of the nodes in the first half of the large network for each sample, according to the ordering
            of the nodes in the large network.
        second_halves_indices
            The indices of the nodes in the second half of the large network for each sample, according to the ordering

        Returns
        -------
        Properly normalized statistics of subnetworks of an observed network.
        """
        return self._calc_bootstrapped_scalar_feature(first_halves_to_use, first_halves_indices, second_halves_indices,
                                                      lambda n: n * (n - 1) / 2)


class ExWeightNumEdges(Metric):
    """
    Weighted sum of the number of edges, based on exogenous attributes.
    """

    # TODO: Collection doesn't necessarily support __getitem__, find a typing hint of a sized Iterable that does.
    def __init__(self, exogenous_attr: Collection):
        super().__init__(requires_graph=False)
        self.exogenous_attr = exogenous_attr
        self.edge_weights = None
        self._calc_edge_weights()

    @abstractmethod
    def _calc_edge_weights(self):
        ...

    @abstractmethod
    def _get_num_weight_mats(self):
        ...

    def _get_total_feature_count(self):
        return self._get_num_weight_mats()

    def calc_change_score(self, current_network: np.ndarray, indices: tuple):
        sign = -1 if current_network[indices[0], indices[1]] else 1
        return sign * self._handle_indices_to_ignore(self.edge_weights)[:, indices[0], indices[1]]

    def calculate(self, input: np.ndarray):
        # TODO - Since most of our focus is on the `calculate_for_sample` functions,
        #  we can convert all individual calculate(x) functions to calculate_for_sample([x])
        #  and reuse code.
        res = np.einsum('ij,kij->k', input, self.edge_weights)
        if not self._is_directed:
            res = res / 2
        return self._handle_indices_to_ignore(res)

    def calculate_for_sample(self, networks_sample: np.ndarray):
        res = self._numba_calculate_for_sample(networks_sample, self.edge_weights)
        if not self._is_directed:
            res = res / 2
        return self._handle_indices_to_ignore(res)

    @staticmethod
    @njit
    def _numba_calculate_for_sample(networks_sample: np.ndarray, edge_weights: np.ndarray):
        reshaped_edge_weights = edge_weights.reshape(edge_weights.shape[0], -1).astype(np.float64)
        reshaped_networks_sample = networks_sample.reshape(-1, networks_sample.shape[2]).astype(np.float64)

        res = reshaped_edge_weights @ reshaped_networks_sample

        return res

    def calculate_mple_regressors(self, observed_network: np.ndarray, edges_indices_lims: tuple[int]):
        # edge_weights shape is (num_weight_mats, n_nodes, n_nodes), the desired outcome is in the shape: (n_nodes**2 - n_nodes, num_weight_mats)
        num_nodes = len(self.exogenous_attr)
        if self._is_directed:
            Xs = self.edge_weights[:, np.eye(num_nodes) == 0].transpose()
        else:
            up_triangle_indices = np.triu_indices(num_nodes, k=1)
            Xs = self.edge_weights[:, up_triangle_indices[0], up_triangle_indices[1]].transpose()
        return Xs[edges_indices_lims[0]:edges_indices_lims[1], :]


class NumberOfEdgesTypesDirected(Metric):
    """
    A metric that counts how many edges exist between different node types in a directed graph.
    For example -
        A graph with `n` nodes, with an exogenous attribute `type=[A, B]` assigned to each node.
        The metric counts the number of edges between nodes of type A->A, A->B, B->A, B->B of a given graph,
        yielding len(type)**2 features.

    Parameters
    -----------
        exogenous_attr : Collection
            A collection of attributes assigned to each node in a graph with n nodes.

            #TODO - len(exogenous_attr)==n is a requirement, right? Should we assert?

        indices_to_ignore: list
            # TODO - Specify the sorting convention
    """

    def __str__(self):
        return "num_edges_between_types_directed"

    def __init__(self, exogenous_attr: Collection, indices_from_user=None):
        self.exogenous_attr = exogenous_attr

        self.unique_types = sorted(list(set(self.exogenous_attr)))

        self._indices_from_user = indices_from_user.copy() if indices_from_user is not None else None
        super().__init__()

        self._effective_feature_count = self._get_effective_feature_count()
        self._indices_to_ignore_up_to_idx = np.cumsum(self._indices_to_ignore)

        self._is_directed = True

        self.indices_of_types = {}
        for i, t in enumerate(self.exogenous_attr):
            if t not in self.indices_of_types.keys():
                self.indices_of_types[t] = [i]
            else:
                self.indices_of_types[t].append(i)

        self.sorted_type_pairs = get_sorted_type_pairs(self.unique_types)
        self._sorted_type_pairs_indices = {pair: i for i, pair in enumerate(self.sorted_type_pairs)}

        self._calc_edge_type_idx_assignment()

    def _calc_edge_type_idx_assignment(self):
        """
        Each edge is assigned with the index of it's corresponding sorted type pair.

        For example, for an n=4 matrix with types [A, B, A, B], the weight matrix is
            [ AA, AB, AA, AB
              BA, BB, BA, BB
              AA, AB, AA, AB
              BA, BB, BA, BB ]
        but instead of AA, AB, ..., we save the corresponding index of the type pair
        in self._sorted_type_pairs (which is sorted alphabetically.)

        """
        num_nodes = len(self.exogenous_attr)
        self._edge_type_idx_assignment = np.zeros((num_nodes, num_nodes)).astype(int)

        for i in range(num_nodes):
            for j in range(num_nodes):
                type_1 = self.exogenous_attr[i]
                type_2 = self.exogenous_attr[j]

                self._edge_type_idx_assignment[i, j] = self._sorted_type_pairs_indices[(type_1, type_2)]

        self._edge_type_idx_assignment += 1  # Increment by 1 to avoid 0-indexing (the index 0 will be kept for non-existing edges)

    def _get_total_feature_count(self):
        return len(self.unique_types) ** 2

    def calculate(self, input: np.ndarray):
        stats = np.zeros(self._get_effective_feature_count())
        idx = 0
        for i, type_pair in enumerate(self.sorted_type_pairs):
            if self._indices_to_ignore is not None and self._indices_to_ignore[i]:
                continue
            stats[idx] = input[np.ix_(self.indices_of_types[type_pair[0]], self.indices_of_types[type_pair[1]])].sum()
            idx += 1
        return stats

    def calculate_for_sample(self, networks_sample: np.ndarray):
        sample_size = networks_sample.shape[2]
        stats = np.zeros((self._get_effective_feature_count(), sample_size))
        idx = 0
        for i, type_pair in enumerate(self.sorted_type_pairs):
            if self._indices_to_ignore is not None and self._indices_to_ignore[i]:
                continue
            stats[idx] = networks_sample[
                np.ix_(self.indices_of_types[type_pair[0]], self.indices_of_types[type_pair[1]])].sum(axis=(0, 1))
            idx += 1
        return stats

    def update_indices_to_ignore(self, indices_to_ignore):
        super().update_indices_to_ignore(indices_to_ignore)
        self._effective_feature_count = self._get_effective_feature_count()
        self._indices_to_ignore_up_to_idx = np.cumsum(self._indices_to_ignore)

    def calc_change_score(self, current_network: np.ndarray, indices: tuple):
        edge_type_pair = (self.exogenous_attr[indices[0]], self.exogenous_attr[indices[1]])
        idx_in_features_vec = self._sorted_type_pairs_indices[edge_type_pair]

        sign = -1 if current_network[indices[0], indices[1]] else 1
        
        change_score = np.zeros(self._effective_feature_count)
        if self._indices_to_ignore is not None:
            if self._indices_to_ignore[idx_in_features_vec]:
                return change_score
            
            change_score[idx_in_features_vec - self._indices_to_ignore_up_to_idx[idx_in_features_vec]] = sign

            return change_score
        else:
            change_score[idx_in_features_vec] = sign
            return change_score

    def calculate_change_score_full_network(self, current_network: np.ndarray):
        """
        Calculate the change score vectors for every edge in the network.

        Parameters
        ----------
        current_network: np.ndarray
            Input network of size (n, n)

        Returns
        -------
        np.ndarray
            Change score matrix of size (n^2-n, metric_num_features) which denotes the change score vector for every edge in the network
        """
        num_nodes = current_network.shape[0]
        num_edges = num_nodes * num_nodes - num_nodes

        change_scores = np.zeros((num_edges, len(self.sorted_type_pairs)))

        sign_matrix = np.where(current_network == 1, -1,
                               1)  # -1 for edges that exist (because we want to remove them), 1 for edges that don't exist (because we want to add them).
        nondiag_signs = sign_matrix[~np.eye(sign_matrix.shape[0], dtype=bool)].flatten()

        # -1 to ignore the empty edge type (edge_weights assigns 0 to empty edges, 1 for the first type-pair, etc...)
        nondiag_weights = self._edge_type_idx_assignment[
                              ~np.eye(self._edge_type_idx_assignment.shape[0], dtype=bool)].flatten() - 1
        change_scores[np.arange(num_edges), nondiag_weights] = nondiag_signs
        change_scores = change_scores.astype(np.int8)

        return self._handle_indices_to_ignore(change_scores, axis=1)

    def calculate_mple_regressors(self, observed_network: np.ndarray, edges_indices_lims: tuple[int]):
        num_edges_to_take = edges_indices_lims[1] - edges_indices_lims[0]
        Xs = np.zeros((num_edges_to_take, len(self.sorted_type_pairs)))
        nondiag_type_idx = self._edge_type_idx_assignment[
                               ~np.eye(self._edge_type_idx_assignment.shape[0], dtype=bool)].flatten() - 1
        Xs[np.arange(num_edges_to_take), nondiag_type_idx[edges_indices_lims[0]:edges_indices_lims[1]]] = 1
        Xs = Xs.astype(np.int8)
        return self._handle_indices_to_ignore(Xs, axis=1)

    def _get_metric_names(self):
        parameter_names = tuple()

        metric_name = str(self)

        for i in range(self._get_total_feature_count()):
            if self._indices_to_ignore is not None and self._indices_to_ignore[i]:
                continue
            type_pair = str(self.sorted_type_pairs[i][0]) + "__" + str(self.sorted_type_pairs[i][1])
            parameter_names += (f"{metric_name}_{type_pair}",)

        return parameter_names

    def _get_ignored_features(self):
        if self._indices_to_ignore is None or not np.any(self._indices_to_ignore):
            return tuple()

        metric_name = str(self)
        ignored_features = ()
        for i in range(self._get_total_feature_count()):
            if self._indices_to_ignore is not None and self._indices_to_ignore[i]:
                type_pair = str(self.sorted_type_pairs[i][0]) + "__" + str(self.sorted_type_pairs[i][1])
                ignored_features += (f"{metric_name}_{type_pair}",)

        return ignored_features


class NodeAttrSum(ExWeightNumEdges):
    def __init__(self, exogenous_attr: Collection, is_directed: bool):
        super().__init__(exogenous_attr)
        self._is_directed = is_directed

    def _calc_edge_weights(self):
        num_nodes = len(self.exogenous_attr)
        self.edge_weights = np.zeros((self._get_num_weight_mats(), num_nodes, num_nodes))
        for i in range(num_nodes):
            for j in range(num_nodes):
                if i == j:
                    continue
                self.edge_weights[0, i, j] = self.exogenous_attr[i] + self.exogenous_attr[j]

    def _get_num_weight_mats(self):
        return 1

    def __str__(self):
        return "node_attribute_sum"


class NodeAttrSumOut(ExWeightNumEdges):
    def __init__(self, exogenous_attr: Collection):
        super().__init__(exogenous_attr)
        self._is_directed = True

    def _calc_edge_weights(self):
        num_nodes = len(self.exogenous_attr)
        self.edge_weights = np.zeros((self._get_num_weight_mats(), num_nodes, num_nodes))
        for i in range(num_nodes):
            self.edge_weights[0, i, :] = self.exogenous_attr[i] * np.ones(num_nodes)
            self.edge_weights[0, i, i] = 0

    def _get_num_weight_mats(self):
        return 1

    def __str__(self):
        return "node_attribute_sum_out"


class NodeAttrSumIn(ExWeightNumEdges):
    def __init__(self, exogenous_attr: Collection):
        super().__init__(exogenous_attr)
        self._is_directed = True

    def _calc_edge_weights(self):
        num_nodes = len(self.exogenous_attr)
        self.edge_weights = np.zeros((self._get_num_weight_mats(), num_nodes, num_nodes))
        for j in range(num_nodes):
            self.edge_weights[0, :, j] = self.exogenous_attr[j] * np.ones(num_nodes)
            self.edge_weights[0, j, j] = 0

    def _get_num_weight_mats(self):
        return 1

    def __str__(self):
        return "node_attribute_in"


class SumDistancesConnectedNeurons(ExWeightNumEdges):
    """
    This class calculates the sum of Euclidean distances between all pairs of connected neurons.
    rows = samples, columns = axes
    """

    def __init__(self, exogenous_attr: pd.DataFrame | pd.Series | np.ndarray | list | tuple, is_directed: bool):
        # todo: decorator that checks inputs and returns np.ndarray if the inputs were valid, and an error otherwise
        if isinstance(exogenous_attr, (pd.DataFrame, pd.Series, list, tuple)):
            exogenous_attr = np.array(exogenous_attr)
        elif not isinstance(exogenous_attr, np.ndarray):
            raise ValueError(
                f"Unsupported type of positions: {type(exogenous_attr)}. Supported types are pd.DataFrame, "
                f"pd.Series, list, tuple and np.ndarray.")
        if len(exogenous_attr.shape) == 1:
            exogenous_attr = exogenous_attr.reshape(-1, 1)

        super().__init__(exogenous_attr)
        self._is_directed = is_directed

    def _calc_edge_weights(self):
        num_nodes = len(self.exogenous_attr)
        self.edge_weights = np.reshape(squareform(pdist(self.exogenous_attr, metric='euclidean')),
                                       (self._get_num_weight_mats(), num_nodes, num_nodes))

    def calc_change_score(self, current_network: np.ndarray, indices: tuple):
        sign = -1 if current_network[indices[0], indices[1]] else 1
        return sign * self.edge_weights[:, indices[0], indices[1]]

    def _get_num_weight_mats(self):
        return 1

    def __str__(self):
        return "sum_distances_connected_neurons"


class NumberOfNodesPerType(Metric):
    def __str__(self):
        return "num_nodes_per_type"

    def __init__(self, metric_node_feature, n_node_categories, feature_dim=1):
        super().__init__(requires_graph=False, metric_type='node', metric_node_feature=metric_node_feature)
        self._is_dyadic_independent = True
        self.n_node_categories = n_node_categories
        self.feature_dim = feature_dim

    def _get_total_feature_count(self):
        """
        How many features does this metric produce, including the ignored ones.
        """
        return self.n_node_categories - 1

    def calculate(self, V: np.ndarray):
        V = V.astype(int)
        if V.shape[1] != 1:
            raise ValueError("the metric NumberOfNodesPerType only works for one kind of node feature.")
        else:
            V = V[:, 0]
        return self._handle_indices_to_ignore(np.bincount(V, minlength=self.n_node_categories)[:-1])  # last category is redundant

    def calc_change_score(self, current_node_features: np.ndarray, idx: int, new_category: int, feature_to_flip=None):
        if current_node_features.shape[1] != 1:
            raise ValueError("the metric NumberOfNodesPerType only works for one kind of node feature.")
        else:
            current_node_features = current_node_features[:, 0]
        old_category = current_node_features[idx]
        changes = np.zeros(self.n_node_categories)
        changes[old_category] = -1
        changes[new_category] = 1
        return self._handle_indices_to_ignore(changes[:-1]) # last category is redundant

    def calculate_for_sample(self, networks_sample: np.ndarray | torch.Tensor):
        """
        We use a trick here to make the bincount operation vectorized: we add each column of the sample
        (namely, features of a different network) a distinct offset, then flatten the sample. Now, we can
        perform bincount on the flattened sample, and reshape it to get back to the counts per network.
        """
        networks_sample = networks_sample.astype(int)
        if networks_sample.shape[1] != 1:
            raise ValueError("the metric NumberOfNodesPerType only works for one kind of node feature.")
        else:
            networks_sample = networks_sample[:, 0]

        n_samples = networks_sample.shape[-1]
        offsets = np.arange(n_samples)[None, :] * self.n_node_categories
        flat_indices = networks_sample + offsets
        flat_counts = np.bincount(flat_indices.ravel(), minlength=n_samples*self.n_node_categories)
        counts_for_sample = flat_counts.reshape(n_samples, self.n_node_categories).T

        return self._handle_indices_to_ignore(counts_for_sample[:-1]) # last category is redundant


class MetricsCollection:

    def __init__(self,
                 metrics: Collection[Metric],
                 is_directed: bool,
                 n_nodes: int,
                 fix_collinearity=True,
                 use_sparse_matrix=False,
                 collinearity_fixer_sample_size=1000,
                 is_collinearity_distributed=False,
                 # TODO: For tests only, find a better solution
                 do_copy_metrics=True,
                 **kwargs):

        if not do_copy_metrics:
            self.metrics = tuple([metric for metric in metrics])
        else:
            self.metrics = tuple([deepcopy(metric) for metric in metrics])

        for m in self.metrics:
            m._n_nodes = n_nodes
            m.initialize_indices_to_ignore()

        self.is_directed = is_directed
        for x in self.metrics:
            if (x._is_directed is not None) and (x._is_directed != self.is_directed):
                model_is_directed_str = "a directed" if self.is_directed else "an undirected"
                metric_is_directed_str = "a directed" if x._is_directed else "an undirected"
                raise ValueError(f"Trying to initialize {model_is_directed_str} model with {metric_is_directed_str} "
                                 f"metric `{str(x)}`!")
        self.n_nodes = n_nodes
        node_feature_name_to_dim = {m.metric_node_feature: m.feature_dim for m in self.metrics if m._metric_type=='node'}
        node_features_dims = list(node_feature_name_to_dim.values())
        node_features_indices_limits = np.cumsum(np.concatenate([[0], node_features_dims])).astype(int)
        node_features_indices = [np.arange(node_features_indices_limits[i], node_features_indices_limits[i+1])
                                 for i in range(len(node_features_indices_limits) - 1)]
        node_feature_names = list(node_feature_name_to_dim.keys())
        self.node_feature_names = dict(zip(node_feature_names, node_features_indices)) # a dict with keys of node feature names and values of lists of indices
        self.n_node_features = sum([len(v) for v in self.node_feature_names.values()])
        self.node_features_n_categories = {m.metric_node_feature: m.n_node_categories for m in self.metrics if m._metric_type=='node'}

        self.use_sparse_matrix = use_sparse_matrix
        self.requires_graph = any([x.requires_graph for x in self.metrics])

        # Returns the number of features that are being calculated. Since a single metric might return more than one
        # feature, the length of the statistics vector might be larger than the amount of metrics. Since it also depends
        # on the network size, n is a mandatory parameters. That's why we're using the get_effective_feature_count
        # function
        self.num_of_features = self.calc_num_of_features()
        self.features_per_metric = np.array([metric._get_effective_feature_count() for metric in self.metrics])

        self._fix_collinearity = fix_collinearity
        self.collinearity_fixer_sample_size = collinearity_fixer_sample_size
        if self._fix_collinearity:
            self.collinearity_fixer(sample_size=self.collinearity_fixer_sample_size,
                                    is_distributed=is_collinearity_distributed,
                                    num_samples_per_job=kwargs.get('num_samples_per_job_collinearity_fixer', 5))

        
        

        self.num_of_metrics = len(self.metrics)
        self.metric_names = tuple([str(metric) for metric in self.metrics])
        self._has_dyadic_dependent_metrics = any([not x._is_dyadic_independent for x in self.metrics])

    def _delete_metric(self, metric: Metric):
        self.metrics = tuple([m for m in self.metrics if m != metric])
        self.requires_graph = any([x.requires_graph for x in self.metrics])

    def calc_num_of_features(self):
        return sum([metric._get_effective_feature_count() for metric in self.metrics])

    def get_metric(self, metric_name: str) -> Metric:
        """
        Get a metric instance
        """
        return self.metrics[self.metric_names.index(metric_name)]

    def get_metric_by_feat_idx(self, idx: int):
        cum_sum_num_feats = 0
        for m in self.metrics:
            cum_sum_num_feats += m._get_effective_feature_count()
            if cum_sum_num_feats > idx:
                return m

    def get_feature_idx_within_metric(self, idx: int):
        cum_sum_num_feats = 0
        for m_idx in range(len(self.metrics)):
            next_met_num_feats = self.metrics[m_idx]._get_effective_feature_count()
            if cum_sum_num_feats + next_met_num_feats > idx:
                # We want to return the index in the "full" array, namely regardless of ignored features. So, in case
                # there are indices that are ignored, we must take the returned index from the array of non-ignored
                # indices, to compensate for the ones ignored by the indexing of MetricCollection (which holds only
                # effective features, after ignoring).
                # For example - if we ignore the degree of the first node, and now want to ignore the degree of the
                # fifth node as well, we must return 4 for the metric to correctly ignore it. But it is the fourth
                # feature corresponding to the metric in the MetricCollection vector of features (not fifth, because the
                # first is missing, not returned by the Metric as it's ignored). By returning the fourth element from
                # the list of non-ignored indices, which is [1, 2, ..., n], we solve the problem.

                effective_idx_within_metric = idx - cum_sum_num_feats
                if hasattr(self.metrics[m_idx], "_indices_to_ignore"):
                    non_ignored_indices = np.where(~self.metrics[m_idx]._indices_to_ignore)[0]
                    # Return the index of the feature with relation to the whole set of features (without ignoring).
                    return non_ignored_indices[effective_idx_within_metric]
                else:
                    return effective_idx_within_metric
            else:
                cum_sum_num_feats += next_met_num_feats

    def calc_statistics_for_binomial_tensor_local(self, tensor_size, p=0.5):
        sample = generate_binomial_tensor(self.n_nodes, self.n_node_features, tensor_size, p=p)

        # Symmetrize samples if not directed
        if not self.is_directed:
            sample[:self.n_nodes, :self.n_nodes] = np.round((sample[:self.n_nodes, :self.n_nodes] + sample[:self.n_nodes, :self.n_nodes].transpose(1, 0, 2)) / 2)

        # Make sure the main diagonal is 0
        sample[np.arange(self.n_nodes, dtype=int), np.arange(self.n_nodes, dtype=int), :] = 0

        # Calculate the features of the sample
        return self.calculate_sample_statistics(sample)

    def calc_statistics_for_binomial_tensor_distributed(self, tensor_size, p=0.5, num_samples_per_job=1):
        # TODO: currently, if tensor_size % num_samples_per_job != 0 the sample size will be larger than tensor_size
        #  specified by the user (it is (tensor_size // num_samples_per_job + 1) * num_samples_per_job). We can pass
        #  tensor_size as an additional argument to children jobs, and validate in
        #  sample_statistics_distributed_calcs.py that we don't calculate for too many networks.
        out_dir_path = (Path.cwd().parent / "OptimizationIntermediateCalculations").resolve()
        data_path = (out_dir_path / "data").resolve()
        os.makedirs(data_path, exist_ok=True)

        with open((data_path / 'metric_collection.pkl').resolve(), 'wb') as f:
            pickle.dump(self, f)

        cmd_line_for_bsub = (f'python ./sample_statistics_distributed_calcs.py '
                             f'--out_dir_path {out_dir_path} '
                             f'--num_samples_per_job {num_samples_per_job} '
                             f'--p {p}')

        num_jobs = int(np.ceil(tensor_size / num_samples_per_job))
        job_array_ids = run_distributed_children_jobs(out_dir_path, cmd_line_for_bsub,
                                                      "distributed_binomial_tensor_statistics.sh",
                                                      num_jobs, "sample_stats")

        # Wait for all jobs to finish. Check the hessian path because it is the last to be computed for each data chunk.
        sample_stats_path = (out_dir_path / "sample_statistics").resolve()
        os.makedirs(sample_stats_path, exist_ok=True)
        wait_for_distributed_children_outputs(num_jobs, sample_stats_path, job_array_ids, "sample_stats")

        # Clean current scripts and data
        shutil.rmtree(data_path)
        shutil.rmtree((out_dir_path / "scripts").resolve())

        # Aggregate results
        whole_sample_statistics = cat_children_jobs_outputs(num_jobs, sample_stats_path, axis=1)

        # clean outputs
        shutil.rmtree(sample_stats_path)
        return whole_sample_statistics

    def remove_feature_by_idx(self, idx: int):
        metric_of_feat = self.get_metric_by_feat_idx(idx)
        is_trimmable = metric_of_feat._get_effective_feature_count() > 1
        if not is_trimmable:
            print(f"Removing the metric {str(metric_of_feat)} from the collection")
            sys.stdout.flush()
            self._delete_metric(metric=metric_of_feat)
        else:
            idx_to_delete_within_metric = self.get_feature_idx_within_metric(idx)
            print(f"Removing the {idx_to_delete_within_metric} feature of {str(metric_of_feat)}")
            sys.stdout.flush()
            metric_of_feat.update_indices_to_ignore([idx_to_delete_within_metric])
        
        self.num_of_features = self.calc_num_of_features()

        # Re-calculate the number of features per metric after deleting a feature.
        self.features_per_metric = np.array([metric._get_effective_feature_count() for metric in self.metrics])        

    def collinearity_fixer(self, sample_size=1000, nonzero_thr=10 ** -1, ratio_threshold=10 ** -6,
                           eigenvec_thr=10 ** -4, is_distributed=False, **kwargs):
        """
        Find collinearity between metrics in the collection.

        Currently this is a naive version that only handles the very simple cases.
        TODO: revisit the threshold and sample size
        """
        is_linearly_dependent = True

        # Sample networks from a maximum entropy distribution, for avoiding edge cases (such as a feature is 0 for
        # all networks in the sample).
        if not is_distributed:
            sample_features = self.calc_statistics_for_binomial_tensor_local(sample_size)
        else:
            sample_features = self.calc_statistics_for_binomial_tensor_distributed(sample_size,
                                                                                   num_samples_per_job=kwargs.get(
                                                                                       "num_samples_per_job", 5))
        while is_linearly_dependent:
            features_cov_mat = sample_features @ sample_features.T

            # Determine whether the matrix of features is invertible. If not - this means there is a non-trivial vector,
            # that when multiplied by the matrix gives the 0 vector. Namely, there is a single set of coefficients that
            # defines a non-trivial linear combination that equals 0, for *all* the sampled feature vectors. This means
            # the features are linearly dependent.
            eigen_vals, eigen_vecs = np.linalg.eigh(features_cov_mat)

            minimal_non_zero_eigen_val = np.min(np.abs(eigen_vals[np.abs(eigen_vals) > nonzero_thr]))
            small_eigen_vals_indices = np.where(np.abs(eigen_vals) / minimal_non_zero_eigen_val < ratio_threshold)[0]

            if small_eigen_vals_indices.size == 0:
                is_linearly_dependent = False
            else:
                print("Collinearity detected, identifying features to remove")
                sys.stdout.flush()

                # For each linear dependency (corresponding to an eigen vector with a low value), mark the indices of
                # features that are involved (identified by a non-zero coefficient in the eigen vector).
                dependent_features_flags = np.zeros((small_eigen_vals_indices.size, self.num_of_features))
                for i in range(small_eigen_vals_indices.size):
                    dependent_features_flags[
                        i, np.where(np.abs(eigen_vecs[:, small_eigen_vals_indices[i]]) > eigenvec_thr)[0]] = 1

                # Calculate the fraction of dependencies each feature is involved in.
                fraction_of_dependencies_involved = dependent_features_flags.mean(axis=0)

                # Sort the features (their indices) by the fraction of dependencies they are involved in (remove first
                # features that are involved in more dependencies). Break ties by the original order of the features in
                # the array (for the consistency of sorting. E.g, if we need to get rid of degree features, always
                # remove them by the order of nodes).
                removal_order = np.lexsort((np.arange(self.num_of_features), -fraction_of_dependencies_involved))

                # Iterate the metrics to find one with multiple features, namely effective number of features that is
                # larger than 1 ('trimmable'). We prefer to trim metrics rather than totally eliminate them from the
                # collection.
                i = 0
                cur_metric = self.get_metric_by_feat_idx(removal_order[i])
                is_trimmable = cur_metric._get_effective_feature_count() > 1
                while (not is_trimmable and i < removal_order.size - 1 and fraction_of_dependencies_involved[
                    removal_order[i]] > 0):
                    i += 1
                    cur_metric = self.get_metric_by_feat_idx(removal_order[i])
                    is_trimmable = cur_metric._get_effective_feature_count() > 1

                # If a trimmable metric was not found (i.e., all features that are involved in the dependency are of
                # metrics with an effective number of features of 1), totally remove the metric that is involved in most
                # dependencies.
                if not is_trimmable:
                    i = 0
                self.remove_feature_by_idx(removal_order[i])

                sample_features = np.delete(sample_features, removal_order[i], axis=0)

    def calculate_statistics(self, W: np.ndarray):
        """
        Calculate the statistics of a graph, formally written as g(y).

        Parameters
        ----------
        W : np.ndarray
            A generalized connectivity matrix: a concatenation of N x N connectivity matrix with an N x k node features matrix.
        
        Returns
        -------
        statistics : np.ndarray
            An array of statistics
        """
        if self.requires_graph:
            G = connectivity_matrix_to_G(W[:self.n_nodes, :self.n_nodes], directed=self.is_directed)

        statistics = np.zeros(self.num_of_features)

        feature_idx = 0
        for metric in self.metrics:
            if metric.requires_graph: # it cannot require graph and also have _metric_type='node'
                input = G
            else:
                if metric._metric_type in ['binary_edge', 'non_binary_edge']:
                    input = W[:self.n_nodes, :self.n_nodes]
                elif metric._metric_type == 'node':
                    feature_indices_to_pass = self.node_feature_names.get(metric.metric_node_feature, list(np.arange(self.n_node_features)))
                    feature_indices_to_pass = [i + self.n_nodes for i in feature_indices_to_pass]
                    input = W[:, feature_indices_to_pass]

            n_features_from_metric = metric._get_effective_feature_count()
            statistics[feature_idx:feature_idx + n_features_from_metric] = metric.calculate(input)
            feature_idx += n_features_from_metric

        return statistics

    def calc_change_scores(self, current_network: np.ndarray, edge_flip_info: dict, node_flip_info={}):
        """
        Calculates the vector of change scores, namely g(net_2) - g(net_1)

        NOTE - this function assumes that the size current_network and self.n_nodes are the same, and doesn't validate
        it, due to runtime considerations. Currently, the only use of this function is within ERGM and
        NaiveMetropolisHastings, so this is fine.
        """
        if self.requires_graph:
            G1 = connectivity_matrix_to_G(current_network[:self.n_nodes, :self.n_nodes], directed=self.is_directed)

        change_scores = np.zeros(self.num_of_features)

        feature_idx = 0
        for i, metric in enumerate(self.metrics):
            # n_features_from_metric = metric._get_effective_feature_count()
            n_features_from_metric = self.features_per_metric[i]

            if metric.requires_graph: # it cannot require graph and also have _metric_type='node'
                input = G1
                change_scores[feature_idx:feature_idx + n_features_from_metric] = metric.calc_change_score(input,
                                                                                                           edge_flip_info['edge'])
            else:
                if metric._metric_type in ['binary_edge', 'non_binary_edge']:
                    input = current_network[:self.n_nodes, :self.n_nodes]
                    change_scores[feature_idx:feature_idx + n_features_from_metric] = metric.calc_change_score(input,
                                                                                                               edge_flip_info['edge'])
                elif metric._metric_type == 'node':
                    feature_indices_to_pass = self.node_feature_names.get(metric.metric_node_feature, list(np.arange(self.n_node_features)))
                    feature_indices_to_pass = [i + self.n_nodes for i in feature_indices_to_pass]
                    if node_flip_info['feature'] not in feature_indices_to_pass:
                        continue
                    input = current_network[:, feature_indices_to_pass]
                    change_scores[feature_idx:feature_idx + n_features_from_metric] = metric.calc_change_score(input,
                                                                                                               current_network[:, feature_indices_to_pass],
                                                                                                               node_flip_info['node'],
                                                                                                               node_flip_info['new_category'])
            feature_idx += n_features_from_metric

        return change_scores

    def calculate_sample_statistics(self, networks_sample: np.ndarray) -> np.ndarray:
        """
        Calculate the statistics over a sample of networks

        Parameters
        ----------
        networks_sample
            The networks sample - an array of n X (n+k) X sample_size
        Returns
        -------
        an array of the statistics vector per sample (num_features X sample_size)
        """
        if networks_sample.shape[0] != self.n_nodes:
            raise ValueError(
                f"Got a sample of networks of size {networks_sample.shape[0]}, but Metrics expect size {self.n_nodes}")

        num_of_samples = networks_sample.shape[2]
        features_of_net_samples = np.zeros((self.num_of_features, num_of_samples))

        if self.requires_graph:
            networks_as_graphs = [connectivity_matrix_to_G(W[:self.n_nodes, :self.n_nodes], self.is_directed) for W in networks_sample]

        if self.use_sparse_matrix:
            networks_as_sparse_tensor = np_tensor_to_sparse_tensor(networks_sample)

        feature_idx = 0
        for metric in self.metrics:
            n_features_from_metric = metric._get_effective_feature_count()

            if metric.requires_graph:
                networks = networks_as_graphs

            elif self.use_sparse_matrix:
                if metric._metric_type in ['binary_edge', 'non_binary_edge']:
                    networks = networks_as_sparse_tensor[:self.n_nodes, :self.n_nodes]
                elif metric._metric_type == 'node':
                    feature_indices_to_pass = self.node_feature_names.get(metric.metric_node_feature, list(np.arange(self.n_node_features)))
                    feature_indices_to_pass = [i + self.n_nodes for i in feature_indices_to_pass]
                    networks = networks_as_sparse_tensor[:, feature_indices_to_pass]
            else:
                if metric._metric_type in ['binary_edge', 'non_binary_edge']:
                    networks = networks_sample[:self.n_nodes, :self.n_nodes]
                elif metric._metric_type == 'node':
                    feature_indices_to_pass = self.node_feature_names.get(metric.metric_node_feature, list(np.arange(self.n_node_features)))
                    feature_indices_to_pass = [i + self.n_nodes for i in feature_indices_to_pass]
                    networks = networks_sample[:, feature_indices_to_pass]

            features = metric.calculate_for_sample(networks)

            if isinstance(features, torch.Tensor):
                if features.is_sparse:
                    features = features.to_dense()
                features = features.numpy()

            features_of_net_samples[feature_idx:feature_idx + n_features_from_metric] = features
            feature_idx += n_features_from_metric

        return features_of_net_samples

    def calculate_change_scores_all_edges(self, current_network: np.ndarray):
        """
        Calculates the vector of change scores for every edge in the network.

        Parameters
        ----------
        current_network : np.ndarray
            The network matrix of size (n,n+k).
        
        Returns
        -------
        change_scores : np.ndarray
            A matrix of size (n**2-n, num_of_features), where each row corresponds to the change scores of the i,j-th edges.
        """
        if self.requires_graph:
            G1 = connectivity_matrix_to_G(current_network[:self.n_nodes, :self.n_nodes], directed=self.is_directed)

        n_nodes = current_network.shape[0]
        num_edges = n_nodes * n_nodes - n_nodes
        change_scores = np.zeros((num_edges, self.num_of_features))

        feature_idx = 0
        for metric in self.metrics:
            if metric.requires_graph:
                input = G1
            else:
                if metric._metric_type in ['binary_edge', 'non_binary_edge']:
                    input = current_network[:self.n_nodes, :self.n_nodes]
                elif metric._metric_type == 'node':
                    feature_indices_to_pass = self.node_feature_names.get(metric.metric_node_feature, list(np.arange(self.n_node_features)))
                    feature_indices_to_pass = [i + self.n_nodes for i in feature_indices_to_pass]
                    input = current_network[:, feature_indices_to_pass]

            n_features_from_metric = metric._get_effective_feature_count()

            if hasattr(metric, "calculate_change_score_full_network"):
                change_scores[:,
                feature_idx:feature_idx + n_features_from_metric] = metric.calculate_change_score_full_network(
                    input)
            else:
                edge_idx = 0
                for i in range(n_nodes):
                    for j in range(n_nodes):
                        if i == j:
                            continue
                        indices = (i, j)
                        change_scores[edge_idx,
                        feature_idx:feature_idx + n_features_from_metric] = metric.calc_change_score(input, indices)
                        edge_idx += 1

            feature_idx += n_features_from_metric

        return change_scores

    def prepare_mple_data(self, observed_network: np.ndarray, edges_indices_lims: tuple[int] | None = None):
        if self.requires_graph:
            G1 = connectivity_matrix_to_G(observed_network[:self.n_nodes, :self.n_nodes], directed=self.is_directed)

        n_nodes = observed_network.shape[0]
        if edges_indices_lims is None:
            num_edges_to_take = n_nodes * n_nodes - n_nodes
            if not self.is_directed:
                num_edges_to_take = num_edges_to_take // 2
            edges_indices_lims = (0, num_edges_to_take)
        else:
            num_edges_to_take = edges_indices_lims[1] - edges_indices_lims[0]
        Xs = np.zeros((num_edges_to_take, self.num_of_features))

        feature_idx = 0
        for metric in self.metrics:
            if metric.requires_graph:
                input = G1
            else:
                if metric._metric_type in ['binary_edge', 'non_binary_edge']:
                    input = observed_network[:self.n_nodes, :self.n_nodes]
                elif metric._metric_type == 'node':
                    feature_indices_to_pass = self.node_feature_names.get(metric.metric_node_feature, list(np.arange(self.n_node_features)))
                    feature_indices_to_pass = [i + self.n_nodes for i in feature_indices_to_pass]
                    input = observed_network[:, feature_indices_to_pass]

            n_features_from_metric = metric._get_effective_feature_count()
            cur_regressors = metric.calculate_mple_regressors(input, edges_indices_lims=edges_indices_lims)
            Xs[:, feature_idx:feature_idx + n_features_from_metric] = cur_regressors

            feature_idx += n_features_from_metric

        if self.is_directed:
            ys = input[~np.eye(n_nodes, dtype=bool)].flatten()[
                 edges_indices_lims[0]:edges_indices_lims[1], None]
        else:
            ys = input[np.triu_indices(n_nodes, 1)][edges_indices_lims[0]:edges_indices_lims[1], None]
        return Xs, ys

    def get_parameter_names(self):
        """
        Returns the names of the parameters of the metrics in the collection.
        """
        parameter_names = tuple()

        for metric in self.metrics:
            parameter_names += metric._get_metric_names()

        return parameter_names

    def get_ignored_features(self):
        parameter_names = tuple()

        for metric in self.metrics:
            parameter_names += metric._get_ignored_features()

        return parameter_names

    def bootstrap_observed_features(self, observed_network: np.ndarray, num_subsamples: int = 1000,
                                    splitting_method: str = 'uniform'):
        observed_connectivity_matrix = observed_network[:self.n_nodes, :self.n_nodes]
        observed_net_size = observed_connectivity_matrix.shape[0]
        second_half_size = observed_net_size // 2
        first_half_size = observed_net_size - second_half_size
        first_halves = np.zeros((first_half_size, first_half_size, num_subsamples))
        second_halves = np.zeros((second_half_size, second_half_size, num_subsamples))
        first_halves_indices = np.zeros((first_half_size, num_subsamples), dtype=int)
        second_halves_indices = np.zeros((second_half_size, num_subsamples), dtype=int)
        for i in range(num_subsamples):
            # TODO: currently we simply split randomly into 2 halves, we want to support other methods for sampling half
            #  of the neurons, and specifically sampling half of the neurons of each if there is a metric with types
            #  involved.
            #  NOTE! This will probably imposes to have another field in `Metric` indicating what is the preferred way
            #  for bootstrapping sampling, and `MetricsCollection` will have to decide somehow what to do based on the
            #  values of all metrics.
            #  NOTE! If there are multiple type metrics, we will probably need to sample according to sub-types defined
            #  as the Cartesian product of all types.
            first_half_indices, second_half_indices = split_network_for_bootstrapping(observed_net_size,
                                                                                      first_half_size,
                                                                                      splitting_method=splitting_method)
            first_halves[:, :, i] = observed_connectivity_matrix[first_half_indices, first_half_indices.T]
            second_halves[:, :, i] = observed_connectivity_matrix[second_half_indices, second_half_indices.T]
            first_halves_indices[:, i] = first_half_indices[:, 0]
            second_halves_indices[:, i] = second_half_indices[:, 0]

        bootstrapped_features = np.zeros((self.num_of_features, num_subsamples))
        # TODO: the next code section is copied from `calculate_for_sample`. The difference is that here we use
        #  `first_halves` and `second_halves` instead of `networks_sample` and a different callable for each metric.
        #  Maybe is can be handled in a single method that gets a callable for metrics and **kwargs or something like
        #  that.
        if self.requires_graph:
            first_halves_as_graphs = [connectivity_matrix_to_G(W, self.is_directed) for W in first_halves]
            second_halves_as_graphs = [connectivity_matrix_to_G(W, self.is_directed) for W in second_halves]

        if self.use_sparse_matrix:
            first_halves_as_sparse_tensor = np_tensor_to_sparse_tensor(first_halves)
            second_halves_as_sparse_tensor = np_tensor_to_sparse_tensor(second_halves)

        feature_idx = 0
        for metric in self.metrics:
            n_features_from_metric = metric._get_effective_feature_count()

            if metric.requires_graph:
                first_halves_to_use = first_halves_as_graphs
                second_halves_to_use = second_halves_as_graphs
            elif self.use_sparse_matrix:
                first_halves_to_use = first_halves_as_sparse_tensor
                second_halves_to_use = second_halves_as_sparse_tensor
            else:
                first_halves_to_use = first_halves
                second_halves_to_use = second_halves

            cur_metric_bootstrapped_features = metric.calculate_bootstrapped_features(
                first_halves_to_use, second_halves_to_use,
                first_halves_indices, second_halves_indices)

            if isinstance(cur_metric_bootstrapped_features, torch.Tensor):
                if cur_metric_bootstrapped_features.is_sparse:
                    cur_metric_bootstrapped_features = cur_metric_bootstrapped_features.to_dense()
                cur_metric_bootstrapped_features = cur_metric_bootstrapped_features.numpy()

            bootstrapped_features[feature_idx:feature_idx + n_features_from_metric] = cur_metric_bootstrapped_features
            feature_idx += n_features_from_metric

        return bootstrapped_features


import unittest
import numpy as np
import networkx as nx

from pyERGM.utils import *
from pyERGM.metrics import *
from pyERGM.ergm import ERGM, BruteForceERGM
from pyERGM import sampling


class Test_MetropolisHastings(unittest.TestCase):
    def setUp(self):
        pass

    def test_flip_network_edge(self):
        stats_calculator = MetricsCollection([NumberOfEdgesUndirected()], is_directed=False, n_nodes=4)
        thetas = np.array([np.log(2)])

        # UNDIRECTED VERSION
        sampler = sampling.NaiveMetropolisHastings(thetas=thetas, metrics_collection=stats_calculator)

        test_W = np.array([
            [0., 0., 0., 1.],
            [0., 0., 1., 0.],
            [0., 1., 0., 0.],
            [1., 0., 0., 0.]
        ])

        node_i = 0
        node_j = 2
        W_plus = test_W.copy()
        sampler._flip_network_edge(W_plus, node_i, node_j)

        expected_W = np.array([
            [0., 0., 1., 1.],
            [0., 0., 1., 0.],
            [1., 1., 0., 0.],
            [1., 0., 0., 0.]
        ])

        self.assertTrue(np.all(W_plus == expected_W))

        node_i = 0
        node_j = 3
        W_minus = test_W.copy()
        sampler._flip_network_edge(W_minus, node_i, node_j)

        expected_W = np.array([
            [0., 0., 0., 0.],
            [0., 0., 1., 0.],
            [0., 1., 0., 0.],
            [0., 0., 0., 0.]
        ])

        self.assertTrue(np.all(W_minus == expected_W))

        # DIRECTED VERSION
        stats_calculator = MetricsCollection([NumberOfEdgesDirected()], is_directed=True, n_nodes=2)
        sampler = sampling.NaiveMetropolisHastings(thetas=thetas, metrics_collection=stats_calculator)

        test_W = np.array([[0, 1], [0, 0]])
        node_i = 1
        node_j = 0
        W_plus = test_W.copy()
        sampler._flip_network_edge(W_plus, node_i, node_j)
        expected_W = np.array([[0, 1], [1, 0]])

        self.assertTrue(np.all(W_plus == expected_W))

        node_i = 0
        node_j = 1
        W_minus = test_W.copy()
        sampler._flip_network_edge(W_minus, node_i, node_j)
        expected_W = np.array([[0, 0], [0, 0]])

        self.assertTrue(np.all(W_minus == expected_W))

    def test__calculate_weighted_change_score_undirected_single_variable(self):
        """
        Test the change score calculation for an undirected graph, based on a single variable - num_edges
        """
        stats_calculator = MetricsCollection([NumberOfEdgesUndirected()], is_directed=False, n_nodes=3)

        theta_edges = 0.5
        thetas = np.array([theta_edges])

        sampler = sampling.NaiveMetropolisHastings(thetas=thetas, metrics_collection=stats_calculator)

        current_W = np.array([
            [0, 0, 1],
            [0, 0, 1],
            [1, 1, 0]
        ])

        change_score = sampler._calculate_weighted_change_score(current_W, {'edge': (0, 1)})
        expected_change_score = 1 * theta_edges
        self.assertEqual(change_score, expected_change_score)

    def test__calculate_weighted_change_score_undirected_multiple_variables(self):
        """
        Test the change score calculation for a undirected graph, based on two variables - num_edges & num_triangles
        """
        stats_calculator = MetricsCollection([NumberOfEdgesUndirected(), NumberOfTriangles()], is_directed=False,
                                             n_nodes=3)

        theta_edges = 2
        theta_triangles = 0.5
        thetas = np.array([theta_edges, theta_triangles])

        sampler = sampling.NaiveMetropolisHastings(thetas=thetas, metrics_collection=stats_calculator)

        current_W = np.array([
            [0, 0, 1],
            [0, 0, 1],
            [1, 1, 0]
        ])

        change_score = sampler._calculate_weighted_change_score(current_W, {'edge': (0, 1)})

        changed_edges = 1
        changed_triangles = 1
        expected_change_score = changed_edges * theta_edges + changed_triangles * theta_triangles

        self.assertEqual(change_score, expected_change_score)

    def test__calculate_weighted_change_score_directed(self):
        """
        Test the change score calculation for a directed graph, based on num_edges
        """
        stats_calculator = MetricsCollection([NumberOfEdgesDirected()], is_directed=True, n_nodes=3)

        theta_edges = -1
        thetas = np.array([theta_edges])

        sampler = sampling.NaiveMetropolisHastings(thetas=thetas, metrics_collection=stats_calculator)

        current_W = np.array([
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 0]
        ])

        current_W_2 = np.array([
            [0, 1, 1],
            [1, 0, 1],
            [0, 1, 0]
        ])

        total_change_score = (sampler._calculate_weighted_change_score(current_W, {'edge': (0, 1)}) +
                              sampler._calculate_weighted_change_score(current_W_2, {'edge': (2, 0)}))

        changed_edges = 2
        expected_change_score = changed_edges * theta_edges

        self.assertEqual(total_change_score, expected_change_score)

    def test__calc_edge_influence_on_features(self):
        np.random.seed(5678234)
        n = 10
        graph = nx.fast_gnp_random_graph(n, 0.5, seed=np.random)
        adj_mat = nx.to_numpy_array(graph)
        sampler = sampling.NaiveMetropolisHastings(np.zeros(n),
                                                   MetricsCollection([NumberOfEdgesDirected(), OutDegree()],
                                                                     is_directed=True, n_nodes=n))
        # The contribution of degree is n, and of number of edges is 1, for each edge
        expected_total_influence = (1 + n) * np.ones(n * (n - 1))
        # The collinearity fixed removed the out degree of the first node, so no contribution for its edges.
        expected_total_influence[:n - 1] = 1
        total_influence = sampler._calc_edge_influence_on_features(adj_mat)
        self.assertTrue(np.all(expected_total_influence == total_influence))

    def test_sample_non_uniform_proposals_smoke_test(self):
        np.random.seed(5678234)
        n = 10
        graph = nx.fast_gnp_random_graph(n, 0.5, seed=np.random)
        adj_mat = nx.to_numpy_array(graph)
        sampler = sampling.NaiveMetropolisHastings(np.zeros(n),
                                                   MetricsCollection([NumberOfEdgesDirected(), OutDegree()],
                                                                     is_directed=True, n_nodes=n))
        sampler.sample(initial_state=adj_mat, num_of_nets=10, edge_proposal_method='features_influence__sum')
        sampler.sample(initial_state=adj_mat, num_of_nets=10, edge_proposal_method='features_influence__softmax')
